import abc
CRMOx=str
CRMOe=None
CRMOG=staticmethod
CRMOv=set
CRMOK=list
CRMOb=map
CRMOY=int
CRMOu=open
import logging
import os
from typing import Dict,Optional,Set
from localstack_ext.bootstrap.cpvcs.models import(Commit,CPVCSNode,CPVCSObj,Revision,StateFileRef,Version)
from localstack_ext.bootstrap.cpvcs.utils.common import config_context
LOG=logging.getLogger(__name__)
class CPVCSSerializer(abc.ABC):
 def __init__(self):
  pass
 @abc.abstractmethod
 def store_obj(self,cpvcs_obj:CPVCSObj)->CRMOx:
  pass
 @abc.abstractmethod
 def retrieve_obj(self,key:CRMOx,remote_path:CRMOx=CRMOe)->Optional[CPVCSObj]:
  pass
 @CRMOG
 def _deserialize_state_files(state_files_str:CRMOx)->Set[StateFileRef]:
  if not state_files_str:
   return CRMOv()
  state_files_attrs=state_files_str.split(";")
  state_files:Set[StateFileRef]=CRMOv()
  for state_file_attrs in state_files_attrs:
   instance_attrs=CRMOK(CRMOb(lambda x:x.split(":")[1],state_file_attrs.split(",")))
   state_files.add(StateFileRef(size=CRMOY(instance_attrs[0]),service=instance_attrs[1],region=instance_attrs[2],hash_ref=instance_attrs[3],file_name=instance_attrs[4],rel_path=instance_attrs[5]))
  return state_files
class VersionSerializerTxt(CPVCSSerializer):
 def store_obj(self,cpvcs_obj:CPVCSNode)->CRMOx:
  with CRMOu(os.path.join(config_context.get_ver_obj_store_path(),cpvcs_obj.hash_ref),"w")as fp:
   fp.write(CRMOx(cpvcs_obj))
  return cpvcs_obj.hash_ref
 def retrieve_obj(self,key:CRMOx,remote_path:CRMOx=CRMOe)->Optional[Version]:
  if remote_path:
   file_path=os.path.join(remote_path,key)
  else:
   file_path=os.path.join(config_context.get_ver_obj_store_path(),key)
  if not os.path.isfile(file_path):
   LOG.debug(f"No Version Obj file found in path {file_path}")
   return
  with CRMOu(os.path.join(config_context.get_ver_obj_store_path(),key),"r")as fp:
   lines=CRMOK(CRMOb(lambda line:line.rstrip(),fp.readlines()))
   version_attrs=CRMOK(CRMOb(lambda line:line.split("=")[1],lines))
   state_files=self._deserialize_state_files(version_attrs[8])
   return Version(parent_ptr=version_attrs[0],hash_ref=version_attrs[1],creator=version_attrs[2],comment=version_attrs[3],version_number=CRMOY(version_attrs[4]),active_revision_ptr=version_attrs[5],outgoing_revision_ptrs=CRMOv(version_attrs[6].split(";")),incoming_revision_ptr=version_attrs[7],state_files=state_files)
class RevisionSerializerTxt(CPVCSSerializer):
 def store_obj(self,cpvcs_obj:Revision)->CRMOx:
  with CRMOu(os.path.join(config_context.get_rev_obj_store_path(),cpvcs_obj.hash_ref),"w")as fp:
   fp.write(CRMOx(cpvcs_obj))
  return cpvcs_obj.hash_ref
 def retrieve_obj(self,key:CRMOx,remote_path:CRMOx=CRMOe)->Optional[Revision]:
  file_path=os.path.join(config_context.get_rev_obj_store_path(),key)
  if not os.path.isfile(file_path):
   LOG.debug(f"No Revision Obj file found in path {file_path}")
   return
  def _deserialize_commit(commit_str:CRMOx)->Commit:
   if not commit_str or commit_str=="None":
    return
   commit_attrs=CRMOK(CRMOb(lambda commit_attr:commit_attr.split(":")[1],commit_str.split(",")))
   return Commit(tail_ptr=commit_attrs[0],head_ptr=commit_attrs[1],message=commit_attrs[2],timestamp=commit_attrs[3],delta_log_ptr=commit_attrs[4])
  with CRMOu(os.path.join(config_context.get_rev_obj_store_path(),key))as fp:
   lines=CRMOK(CRMOb(lambda line:line.rstrip(),fp.readlines()))
   revision_attrs=CRMOK(CRMOb(lambda line:line.split("=")[1],lines))
   state_files=self._deserialize_state_files(revision_attrs[5])
   return Revision(parent_ptr=revision_attrs[0],hash_ref=revision_attrs[1],creator=revision_attrs[2],rid=revision_attrs[3],revision_number=CRMOY(revision_attrs[4]),state_files=state_files,assoc_commit=_deserialize_commit(revision_attrs[6]))
version_serializer=VersionSerializerTxt()
revision_serializer=RevisionSerializerTxt()
txt_serializers:Dict[CRMOx,CPVCSSerializer]={"version":version_serializer,"revision":revision_serializer}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.constants import TEST_AWS_ACCOUNT_ID,TRUE_STRINGS
rftAR=staticmethod
rftAe=super
rftAO=classmethod
rftAF=None
rftAH=int
rftAn=list
rftAB=True
rftAE=isinstance
rftAc=str
rftAy=Exception
from localstack.services.cloudformation.service_models import(REF_ATTRS,REF_ID_ATTRS,GenericBaseModel)
from localstack.utils.aws import aws_stack
from localstack_ext.services.cognito.cognito_idp_api import get_issuer_url
from localstack_ext.utils.aws import aws_utils
class CognitoUserPool(GenericBaseModel):
 @rftAR
 def cloudformation_type():
  return "AWS::Cognito::UserPool"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ATTRS:
   return self.get_physical_resource_id(attribute_name)
  pool_id=self._get_id()
  if attribute_name=="Arn":
   return aws_utils.cognito_userpool_arn(pool_id)
  if attribute_name=="ProviderName":
   return "cognito-idp.{r}.amazonaws.com/{r}_{a}".format(r=aws_stack.get_region(),a=TEST_AWS_ACCOUNT_ID)
  if attribute_name=="ProviderURL":
   return get_issuer_url(pool_id=pool_id)
  return rftAe(CognitoUserPool,self).get_cfn_attribute(attribute_name)
 def get_physical_resource_id(self,attribute,**kwargs):
  pool_id=self._get_id()
  if not pool_id:
   return pool_id
  if attribute in REF_ATTRS:
   return pool_id
  return aws_utils.cognito_userpool_arn(pool_id)
 def get_resource_name(self):
  return self.props.get("PoolName")
 def _get_id(self):
  props=self.props
  return props.get("UserPoolId")or props.get("Id")
 @rftAO
 def fetch_details(cls,pool_name):
  client=aws_stack.connect_to_service("cognito-idp")
  pools=client.list_user_pools(MaxResults=100)["UserPools"]
  return([p for p in pools if p["Name"]==pool_name]or[rftAF])[0]
 @rftAR
 def get_deploy_templates():
  def get_user_pool_params(params,**kwargs):
   attr_list=["Policies","LambdaConfig","AutoVerifiedAttributes","AliasAttributes","UsernameAttributes","VerificationMessageTemplate","EmailVerificationMessage","EmailVerificationSubject","SmsVerificationMessage","SmsAuthenticationMessage","DeviceConfiguration","EmailConfiguration","SmsConfiguration","UserPoolTags","AdminCreateUserConfig","Schema","UserPoolAddOns"]
   attr_map={attr:attr for attr in attr_list}
   attr_map["PoolName"]="UserPoolName"
   result={k:params.get(v)for k,v in attr_map.items()}
   policies=result.get("Policies")or{}
   pw_policy=policies.get("PasswordPolicy")or{}
   if pw_policy.get("MinimumLength"):
    pw_policy["MinimumLength"]=rftAH(pw_policy["MinimumLength"])
   bool_attrs=["RequireLowercase","RequireNumbers","RequireSymbols","RequireUppercase"]
   true_values=rftAn(TRUE_STRINGS)+[rftAB]
   for bool_attr in bool_attrs:
    if bool_attr in pw_policy:
     pw_policy[bool_attr]=pw_policy[bool_attr]in true_values
   if rftAE(result.get("AutoVerifiedAttributes"),rftAc):
    result["AutoVerifiedAttributes"]=[result["AutoVerifiedAttributes"]]
   for schema_attr in result.get("Schema")or[]:
    attr_constr=schema_attr.get("StringAttributeConstraints")
    if attr_constr:
     attr_constr["MinLength"]=rftAc(attr_constr.get("MinLength",1))
     attr_constr["MaxLength"]=rftAc(attr_constr.get("MaxLength",50))
   return result
  return{"create":{"function":"create_user_pool","parameters":get_user_pool_params}}
class UserPoolGroup(GenericBaseModel):
 @rftAR
 def cloudformation_type():
  return "AWS::Cognito::UserPoolGroup"
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("GroupName")
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("cognito-idp")
  props=self.props
  group_name=self.resolve_refs_recursively(stack_name,props.get("GroupName"),resources)
  pool_id=self.resolve_refs_recursively(stack_name,props.get("UserPoolId"),resources)
  groups=client.list_groups(UserPoolId=pool_id)["Groups"]
  result=[g for g in groups if g["GroupName"]==group_name]
  return(result or[rftAF])[0]
 @rftAR
 def get_deploy_templates():
  return{"create":{"function":"create_group"},"delete":{"function":"delete_group","parameters":["GroupName","UserPoolId"]}}
class IdentityPool(GenericBaseModel):
 @rftAR
 def cloudformation_type():
  return "AWS::Cognito::IdentityPool"
 def get_cfn_attribute(self,attribute_name):
  try:
   return rftAe(IdentityPool,self).get_cfn_attribute(attribute_name)
  except rftAy:
   if attribute_name in REF_ID_ATTRS:
    return self.get_physical_resource_id(attribute_name)
   if attribute_name=="ProviderName":
    return "cognito.{r}.amazonaws.com/{r}_{a}".format(r=aws_stack.get_region(),a=TEST_AWS_ACCOUNT_ID)
   raise
 def get_physical_resource_id(self,attribute,**kwargs):
  if attribute in REF_ID_ATTRS:
   return self.props.get("IdentityPoolId")
 @rftAO
 def fetch_details(cls,pool_name):
  client=aws_stack.connect_to_service("cognito-identity")
  pools=client.list_identity_pools(MaxResults=100)["IdentityPools"]
  result=[p for p in pools if p["IdentityPoolName"]==pool_name]
  return(result or[rftAF])[0]
class CognitoUserPoolClient(GenericBaseModel):
 @rftAR
 def cloudformation_type():
  return "AWS::Cognito::UserPoolClient"
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ID_ATTRS:
   return self.get_physical_resource_id(attribute_name)
  return rftAe(CognitoUserPoolClient,self).get_cfn_attribute(attribute_name)
 def get_physical_resource_id(self,attribute,**kwargs):
  if attribute in REF_ID_ATTRS:
   return self.props.get("ClientId")
 @rftAO
 def fetch_details(cls,pool_id,client_name):
  client=aws_stack.connect_to_service("cognito-idp")
  clients=client.list_user_pool_clients(UserPoolId=pool_id)["UserPoolClients"]
  return([c for c in clients if c["ClientName"]==client_name]or[rftAF])[0]
class CognitoUserPoolDomain(GenericBaseModel):
 @rftAR
 def cloudformation_type():
  return "AWS::Cognito::UserPoolDomain"
 def get_physical_resource_id(self,attribute,**kwargs):
  if attribute in REF_ID_ATTRS:
   return self.props.get("Domain")
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("cognito-idp")
  domain_name=self.resolve_refs_recursively(stack_name,self.props.get("Domain"),resources)
  domain=client.describe_user_pool_domain(Domain=domain_name)["DomainDescription"]
  return domain or rftAF
 @rftAR
 def get_deploy_templates():
  return{"create":{"function":"create_user_pool_domain"},"delete":{"function":"delete_user_pool_domain","parameters":{"UserPoolId":"UserPoolId","Domain":"Domain"}}}
class CognitoIdentityPoolRoleAttachment(GenericBaseModel):
 @rftAR
 def cloudformation_type():
  return "AWS::Cognito::IdentityPoolRoleAttachment"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("cognito-identity")
  pool_id=self.resolve_refs_recursively(stack_name,self.props.get("IdentityPoolId"),resources)
  roles=client.get_identity_pool_roles(IdentityPoolId=pool_id)
  if roles:
   roles["_deployed"]=rftAB
  return roles or rftAF
 def get_physical_resource_id(self,attribute,**kwargs):
  props=self.props
  return props.get("_deployed")and "cognito-pool-roles-%s"%props.get("IdentityPoolId")
 @rftAR
 def get_deploy_templates():
  return{"create":{"function":"set_identity_pool_roles"}}
class CognitoUserPoolIdentityProvider(GenericBaseModel):
 @rftAR
 def cloudformation_type():
  return "AWS::Cognito::UserPoolIdentityProvider"
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("ProviderName")
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("cognito-idp")
  props=self.props
  pool_id=self.resolve_refs_recursively(stack_name,props.get("UserPoolId"),resources)
  prov_name=self.resolve_refs_recursively(stack_name,props.get("ProviderName"),resources)
  providers=client.list_identity_providers(UserPoolId=pool_id)["Providers"]
  provider=[p for p in providers if p["ProviderName"]==prov_name]
  return(provider or[rftAF])[0]
 @rftAR
 def get_deploy_templates():
  return{"create":{"function":"create_identity_provider"},"delete":{"function":"delete_identity_provider","parameters":{"UserPoolId":"UserPoolId","ProviderName":"ProviderName"}}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import GenericBaseModel
wJbvy=staticmethod
wJbvD=None
from localstack.utils.aws import aws_stack
class BackupPlan(GenericBaseModel):
 @wJbvy
 def cloudformation_type():
  return "AWS::Backup::BackupPlan"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("backup")
  plans=client.list_backup_plans().get("BackupPlansList",[])
  plan_name=self.resolve_refs_recursively(stack_name,self.props["BackupPlanName"],resources)
  result=[p for p in plans if p["BackupPlanName"]==plan_name]
  return(result or[wJbvD])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("BackupPlanId")
 @wJbvy
 def get_deploy_templates():
  return{"create":{"function":"create_backup_plan"},"delete":{"function":"delete_backup_plan","parameters":["BackupPlanId"]}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.utils.aws import aws_models
KfESW=super
KfESh=None
KfESL=id
class LambdaLayer(aws_models.LambdaFunction):
 def __init__(self,arn):
  KfESW(LambdaLayer,self).__init__(arn)
  self.cwd=KfESh
  self.runtime=""
  self.handler=""
  self.envvars={}
  self.versions={}
class BaseComponent(aws_models.Component):
 def name(self):
  return self.KfESL.split(":")[-1]
class RDSDatabase(BaseComponent):
 def __init__(self,KfESL,env=KfESh):
  KfESW(RDSDatabase,self).__init__(KfESL,env=env)
class RDSCluster(BaseComponent):
 def __init__(self,KfESL,env=KfESh):
  KfESW(RDSCluster,self).__init__(KfESL,env=env)
class AppSyncAPI(BaseComponent):
 def __init__(self,KfESL,env=KfESh):
  KfESW(AppSyncAPI,self).__init__(KfESL,env=env)
class AmplifyApp(BaseComponent):
 def __init__(self,KfESL,env=KfESh):
  KfESW(AmplifyApp,self).__init__(KfESL,env=env)
class ElastiCacheCluster(BaseComponent):
 def __init__(self,KfESL,env=KfESh):
  KfESW(ElastiCacheCluster,self).__init__(KfESL,env=env)
class TransferServer(BaseComponent):
 def __init__(self,KfESL,env=KfESh):
  KfESW(TransferServer,self).__init__(KfESL,env=env)
class CloudFrontDistribution(BaseComponent):
 def __init__(self,KfESL,env=KfESh):
  KfESW(CloudFrontDistribution,self).__init__(KfESL,env=env)
class CodeCommitRepository(BaseComponent):
 def __init__(self,KfESL,env=KfESh):
  KfESW(CodeCommitRepository,self).__init__(KfESL,env=env)
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import logging
cbfVn=True
cbfVk=False
cbfVt=None
cbfVX=Exception
cbfVC=open
cbfVu=bool
cbfVe=str
cbfVo=int
cbfVz=dict
cbfVE=list
cbfVx=isinstance
cbfVJ=super
import os
import sys
from typing import Any,Callable,Dict,List
import click
from localstack.cli import LocalstackCli,LocalstackCliPlugin,console
class ProCliPlugin(LocalstackCliPlugin):
 name="pro"
 def should_load(self):
  e=os.getenv("LOCALSTACK_API_KEY")
  return cbfVn if e else cbfVk
 def is_active(self):
  return self.should_load()
 def attach(self,cli:LocalstackCli)->cbfVt:
  group:click.Group=cli.group
  group.add_command(cmd_login)
  group.add_command(cmd_logout)
  group.add_command(daemons)
  group.add_command(pod)
  group.add_command(dns)
  group.add_command(cpvcs)
@click.group(name="daemons",help="Manage local daemon processes")
def daemons():
 pass
@click.command(name="login",help="Log in with your account credentials")
@click.option("--username",help="Username for login")
@click.option("--provider",default="internal",help="OAuth provider (default: localstack internal login)")
def cmd_login(username,provider):
 from localstack_ext.bootstrap import auth
 try:
  auth.login(provider,username)
  console.print("successfully logged in")
 except cbfVX as e:
  console.print("authentication error: %s"%e)
@click.command(name="logout",help="Log out and delete any session tokens")
def cmd_logout():
 from localstack_ext.bootstrap import auth
 try:
  auth.logout()
  console.print("successfully logged out")
 except cbfVX as e:
  console.print("logout error: %s"%e)
@daemons.command(name="start",help="Start local daemon processes")
def cmd_daemons_start():
 from localstack_ext.bootstrap import local_daemon
 console.log("Starting local daemons processes ...")
 local_daemon.start_in_background()
@daemons.command(name="stop",help="Stop local daemon processes")
def cmd_daemons_stop():
 from localstack_ext.bootstrap import local_daemon
 console.log("Stopping local daemons processes ...")
 local_daemon.kill_servers()
@daemons.command(name="log",help="Show log of daemon process")
def cmd_daemons_log():
 from localstack_ext.bootstrap import local_daemon
 file_path=local_daemon.get_log_file_path()
 if not os.path.isfile(file_path):
  console.print("no log found")
 else:
  with cbfVC(file_path,"r")as fd:
   for line in fd:
    sys.stdout.write(line)
    sys.stdout.flush()
@click.group(name="dns",help="Manage DNS settings of your host")
def dns():
 pass
@dns.command(name="systemd-resolved",help="Manage DNS settings of systemd-resolved (Ubuntu, Debian etc.)")
@click.option("--revert",is_flag=cbfVn,help="Revert systemd-resolved settings for the docker interface")
def cmd_dns_systemd(revert:cbfVu):
 import localstack_ext.services.dns_server
 from localstack_ext.bootstrap.dns_utils import configure_systemd
 console.print("Configuring systemd-resolved...")
 logger_name=localstack_ext.services.dns_server.LOG.name
 localstack_ext.services.dns_server.LOG=ConsoleLogger(logger_name)
 configure_systemd(revert)
def _cpvcs_initialized(pod_name:cbfVe)->cbfVu:
 from localstack_ext.bootstrap.cpvcs.utils.common import config_context
 config_context.set_pod_context(pod_name=pod_name)
 if not config_context.is_initialized():
  console.print("[red]Error:[/red] Could not find local CPVCS instance")
  return cbfVk
 return cbfVn
@click.group(name="cpvcs",help="Experimental Cloud Pods with elaborate versioning mechanism")
def cpvcs():
 from localstack_ext.bootstrap.licensing import is_logged_in
 if not is_logged_in():
  console.print("[red]Error:[/red] not logged in, please log in first")
  sys.exit(1)
@cpvcs.command(name="init",help="Creates a new cloud pod with cpvcs enabled")
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
def cmd_cpvcs_init(name:cbfVe):
 from localstack_ext.bootstrap import pods_client
 from localstack_ext.bootstrap.cpvcs.utils.common import config_context
 config_context.set_pod_context(name)
 if config_context.is_initialized():
  console.print(f"[red]Error:[/red] CPVCS already instanciated for pod {name}")
 else:
  pods_client.init_cpvcs(pod_name=name,pre_config={"backend":"cpvcs"})
  console.print("Successfully created local CPVCS instance!")
@cpvcs.command(name="delete",help="Deletes the specified cloud pod. By default only locally")
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
@click.option("-r","--remote",help="Whether the Pod should also be deleted remotely.",is_flag=cbfVn,default=cbfVk)
def cmd_cpvcs_delete(name:cbfVe,remote:cbfVu):
 from localstack_ext.bootstrap import pods_client
 result=pods_client.delete_pod(pod_name=name,remote=remote,pre_config={"backend":"cpvcs"})
 if result:
  console.print(f"Successfully deleted {name}")
 else:
  console.print(f"[yellow]{name} not available locally[/yellow]")
@cpvcs.command(name="register",help="Registers a local cloudpod instance with platform")
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
def cmd_cpvcs_register_remote(name:cbfVe):
 from localstack_ext.bootstrap import pods_client
 if _cpvcs_initialized(pod_name=name):
  result=pods_client.register_remote(pod_name=name,pre_config={"backend":"cpvcs"})
  if result:
   console.print(f"Successfully registred {name} with remote!")
  else:
   console.print(f"[red]Error:[/red] Pod with name {name} is already registered")
@cpvcs.command(name="rename",help="Renames the pod. If the pod is remotely registered, change is also propagated to remote")
@click.option("-n","--name",help="Current Name of the cloud pod",required=cbfVn)
@click.option("-nn","--new-name",help="New name of the cloud pod",required=cbfVn)
def cmd_cpvcs_rename(name:cbfVe,new_name:cbfVe):
 from localstack_ext.bootstrap import pods_client
 if _cpvcs_initialized(pod_name=name):
  result=pods_client.rename_pod(current_pod_name=name,new_pod_name=new_name,pre_config={"backend":"cpvcs"})
  if result:
   console.print(f"Successfully renamed {name} to {new_name}")
  else:
   console.print(f"[red]Error:[/red] Failed to rename {name} to {new_name}")
@cpvcs.command(name="commit",help="Commits the current expansion point and creates a new (empty) revision")
@click.option("-m","--message",help="Add a comment describing the revision")
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
def cmd_cpvcs_commit(message:cbfVe,name:cbfVe):
 from localstack_ext.bootstrap import pods_client
 if _cpvcs_initialized(pod_name=name):
  pods_client.commit_state(pod_name=name,pre_config={"backend":"cpvcs"},message=message)
  console.print("Successfully commited the current state")
@cpvcs.command(name="push",help="Creates a new version by using the state files in the current expansion point (latest commit)")
@click.option("--squash",is_flag=cbfVn,help="Squashes commits together, so only the latest commit is stored in the revision graph")
@click.option("--three-way",is_flag=cbfVn,default=cbfVk,help="")
@click.option("-m","--message",help="Add a comment describing the version")
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
def cmd_cpvcs_push(squash:cbfVu,message:cbfVe,name:cbfVe,three_way:cbfVu):
 from localstack_ext.bootstrap import pods_client
 if _cpvcs_initialized(pod_name=name):
  pods_client.push_state(pod_name=name,pre_config={"backend":"cpvcs"},squash_commits=squash,comment=message)
  console.print("Successfully pushed the current state")
@cpvcs.command(name="push-overwrite",help="Overwrites a version with the content from the latest commit of the currently selected version")
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
@click.option("-v","--version",type=cbfVo)
@click.option("-m","--message",required=cbfVk)
def cmd_cpvcs_push_overwrite(version:cbfVo,message:cbfVe,name:cbfVe):
 from localstack_ext.bootstrap import pods_client
 if _cpvcs_initialized(pod_name=name):
  result=pods_client.push_overwrite(version=version,pod_name=name,comment=message,pre_config={"backend":"cpvcs"})
  if result:
   console.print("Successfully overwritten state of version ")
@cpvcs.command(name="inject",help="Injects the state from a version into the application runtime")
@click.option("-v","--version",default="-1",type=cbfVo,help="Loads the state of the specified version - Most recent one by default")
@click.option("--reset",is_flag=cbfVn,default=cbfVk,help="Will reset the application state before injecting")
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
def cmd_cpvcs_inject(version:cbfVo,reset:cbfVu,name:cbfVe):
 from localstack_ext.bootstrap import pods_client
 if _cpvcs_initialized(pod_name=name):
  pods_client.inject_state(pod_name=name,version=version,reset_state=reset,pre_config={"backend":"cpvcs"})
@click.option("--inject/--no-inject",default=cbfVn,help="Whether the latest version of the pulled pod should be injected")
@click.option("--reset/--no-reset",default=cbfVn,help="Whether the current application state should be reset after the pod has been pulled")
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
@cpvcs.command(name="pull",help="Injects the state from a version into the application runtime")
def cmd_cpvcs_pull(name:cbfVe,inject:cbfVu,reset:cbfVu):
 from localstack_ext.bootstrap import pods_client
 pods_client.pull_state(pod_name=name,inject_version_state=inject,reset_state_before=reset,pre_config={"backend":"cpvcs"})
@cpvcs.command(name="list-pods",help="Lists all pods and indicates which pods exist locally and, by default, which ones are managed remotely")
@click.option("--remote","-r",is_flag=cbfVn,default=cbfVk)
def cmd_cpvcs_list_pods(remote:cbfVu):
 from localstack_ext.bootstrap import pods_client
 pods=pods_client.list_pods_cpvcs(remote=remote,pre_config={"backend":"cpvcs"})
 if not pods:
  console.print(f"[yellow]No pods available {'locally' if not remote else ''}[/yellow]")
 else:
  console.print("\n".join(pods))
@cpvcs.command(name="versions",help="Lists all available version numbers")
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
def cmd_cpvcs_versions(name:cbfVe):
 if _cpvcs_initialized(pod_name=name):
  from localstack_ext.bootstrap import pods_client
  version_list=pods_client.list_versions(pod_name=name,pre_config={"backend":"cpvcs"})
  result="\n".join(version_list)
  console.print(result)
@cpvcs.command(name="version-info")
@click.option("-v","--version",required=cbfVn,type=cbfVo)
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
def cmd_cpvcs_version_info(version:cbfVo,name:cbfVe):
 if _cpvcs_initialized(pod_name=name):
  from localstack_ext.bootstrap import pods_client
  info=pods_client.get_version_info(version=version,pod_name=name,pre_config={"backend":"cpvcs"})
  console.print_json(info)
@cpvcs.command(name="metamodel",help="Displays the content metamodel as json")
@click.option("-v","--version",type=cbfVo,default=-1,help="Latest version by default")
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
def cmd_cpvcs_version_metamodel(version:cbfVo,name:cbfVe):
 if _cpvcs_initialized(pod_name=name):
  from localstack_ext.bootstrap import pods_client
  metamodel=pods_client.get_version_metamodel(version=version,pod_name=name,pre_config={"backend":"cpvcs"})
  if metamodel:
   console.print_json(metamodel)
  else:
   console.print(f"[red]Could not find metaodel for pod {name} with version {version}[/red]")
@cpvcs.command(name="set-version",help="Set HEAD to a specific version")
@click.option("-v","--version",required=cbfVn,type=cbfVo,help="The version the state should be set to")
@click.option("--inject/--no-inject",default=cbfVn,help="Whether the state should be directly injected into the application runtime after changing version")
@click.option("--reset/--no-reset",default=cbfVn,help="Whether the current application state should be reset before changing version")
@click.option("--commit-before",is_flag=cbfVk,help="Whether the current application state should be commited to the currently selected version before changing version")
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
def cmd_cpvcs_set_version(version:cbfVo,inject:cbfVu,reset:cbfVu,commit_before:cbfVu,name:cbfVe):
 if _cpvcs_initialized(pod_name=name):
  from localstack_ext.bootstrap import pods_client
  pods_client.set_version(version=version,inject_version_state=inject,reset_state=reset,commit_before=commit_before,pod_name=name,pre_config={"backend":"cpvcs"})
@cpvcs.command(name="commits",help="Shows the commit history of a version")
@click.option("--version","-v",default=-1)
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
def cmd_cpvcs_commits(version:cbfVo,name:cbfVe):
 if _cpvcs_initialized(pod_name=name):
  from localstack_ext.bootstrap import pods_client
  commits=pods_client.list_version_commits(version=version,pod_name=name,pre_config={"backend":"cpvcs"})
  result="\n".join(commits)
  console.print(result)
@cpvcs.command(name="commit-diff",help="Shows the changes made by a commit")
@click.option("--version","-v",required=cbfVn)
@click.option("--commit","-c",required=cbfVn)
@click.option("-n","--name",help="Name of the cloud pod",required=cbfVn)
def cmd_cpvcs_commit_diff(version:cbfVo,commit:cbfVo,name:cbfVe):
 if _cpvcs_initialized(pod_name=name):
  from localstack_ext.bootstrap import pods_client
  commit_diff=pods_client.get_commit_diff(version=version,commit=commit,pod_name=name,pre_config={"backend":"cpvcs"})
  if commit_diff:
   console.print_json(commit_diff)
  else:
   console.print(f"[red]Error:[/red] Commit {commit} not found for version {version}")
@click.group(name="pod",help="Manage state of local cloud pods")
def pod():
 from localstack_ext.bootstrap.licensing import is_logged_in
 if not is_logged_in():
  console.print("[red]Error:[/red] not logged in, please log in first")
  sys.exit(1)
@pod.command(name="list",help="Get a list of available local cloud pods")
def cmd_pod_list():
 status=console.status("Fetching list of pods from server ...")
 status.start()
 from localstack import config
 from localstack.utils.common import format_bytes
 from localstack_ext.bootstrap import pods_client
 try:
  result=pods_client.list_pods(cbfVt)
  status.stop()
  columns={"pod_name":"Name","backend":"Backend","url":"URL","size":"Size","state":"State"}
  print_table(columns,result,formatters={"size":format_bytes})
 except cbfVX as e:
  status.stop()
  if config.DEBUG:
   console.print_exception()
  else:
   console.print("[red]Error:[/red]",e)
@pod.command(name="create",help="Create a new local cloud pod")
def cmd_pod_create():
 msg="Please head over to https://app.localstack.cloud to create a new cloud pod. (CLI support is coming soon)"
 console.print(msg)
@pod.command(name="push",help="Push the state of the LocalStack instance to a cloud pod")
@click.argument("name")
def cmd_pod_push(name:cbfVe):
 from localstack_ext.bootstrap import pods_client
 pods_client.push_state(name)
@pod.command(name="pull",help="Pull the state of a cloud pod into the running LocalStack instance")
@click.argument("name")
def cmd_pod_pull(name:cbfVe):
 from localstack_ext.bootstrap import pods_client
 pods_client.pull_state(name)
@pod.command(name="reset",help="Reset the local state to get a fresh LocalStack instance")
def cmd_pod_reset():
 from localstack_ext.bootstrap import pods_client
 pods_client.reset_local_state()
def print_table(columns:Dict[cbfVe,cbfVe],rows:List[Dict[cbfVe,Any]],formatters:Dict[cbfVe,Callable[[Any],cbfVe]]=cbfVt):
 from rich.table import Table
 if formatters is cbfVt:
  formatters=cbfVz()
 t=Table()
 for k,name in columns.items():
  t.add_column(name)
 for row in rows:
  cells=cbfVE()
  for c in columns.keys():
   cell=row.get(c)
   if c in formatters:
    cell=formatters[c](cell)
   if cell is cbfVt:
    cell=""
   if not cbfVx(cell,cbfVe):
    cell=cbfVe(cell)
   cells.append(cell)
  t.add_row(*cells)
 console.print(t)
class ConsoleLogger(logging.Logger):
 def __init__(self,name):
  cbfVJ(ConsoleLogger,self).__init__(name)
 def info(self,msg:Any,*args:Any,**kwargs:Any)->cbfVt:
  console.print(msg%args)
 def warning(self,msg:Any,*args:Any,**kwargs:Any)->cbfVt:
  console.print("[red]Warning:[/red] ",msg%args)
 def error(self,msg:Any,*args:Any,**kwargs:Any)->cbfVt:
  console.print("[red]Error:[/red] ",msg%args)
 def exception(self,msg:Any,*args:Any,**kwargs:Any)->cbfVt:
  console.print("[red]Error:[/red] ",msg%args)
  console.print_exception()
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack_ext import __version__

# version of localstack-ext
# TODO deprecated - use localstack-ext.__version__ instead
VERSION = __version__

# default expiry seconds for Cognito access tokens
TOKEN_EXPIRY_SECONDS = 24 * 60 * 60

# name of Docker registry for Lambda images
DEFAULT_LAMBDA_DOCKER_REGISTRY = "localstack/lambda"

# Github repo with various code artifacts
ARTIFACTS_REPO = "https://github.com/localstack/localstack-artifacts"

# request path for local pod management API
API_PATH_PODS = "/_pods"



import json
rxSqH=staticmethod
rxSqW=None
rxSqA=classmethod
rxSqB=super
rxSqP=type
from localstack.services.cloudformation.service_models import REF_ID_ATTRS,GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.common import select_attributes,short_uid
class ApiGatewayV2VpcLink(GenericBaseModel):
 @rxSqH
 def cloudformation_type():
  return "AWS::ApiGatewayV2::VpcLink"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("apigatewayv2")
  links=client.get_vpc_links()["Items"]
  link_name=self.resolve_refs_recursively(stack_name,self.props["Name"],resources)
  result=[e for e in links if e["Name"]==link_name]
  return(result or[rxSqW])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("VpcLinkId")
 @rxSqH
 def get_deploy_templates():
  return{"create":{"function":"create_vpc_link"},"delete":{"function":"delete_vpc_link","parameters":["VpcLinkId"]}}
class ApiGatewayV2DomainName(GenericBaseModel):
 @rxSqH
 def cloudformation_type():
  return "AWS::ApiGatewayV2::DomainName"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("apigatewayv2")
  apis=client.get_domain_names()["Items"]
  domain_name=self.resolve_refs_recursively(stack_name,self.props["DomainName"],resources)
  result=([a for a in apis if a["DomainName"]==domain_name]or[rxSqW])[0]
  return result
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("DomainName")
 @rxSqH
 def get_deploy_templates():
  return{"create":{"function":"create_domain_name"},"delete":{"function":"delete_domain_name","parameters":["DomainName"]}}
class ApiGatewayV2Authorizer(GenericBaseModel):
 @rxSqH
 def cloudformation_type():
  return "AWS::ApiGatewayV2::Authorizer"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("apigatewayv2")
  props=self.props
  api_id=self.resolve_refs_recursively(stack_name,props["ApiId"],resources)
  auth_name=self.resolve_refs_recursively(stack_name,props["Name"],resources)
  apis=client.get_authorizers(ApiId=api_id)["Items"]
  result=([a for a in apis if a["Name"]==auth_name]or[rxSqW])[0]
  return result
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("AuthorizerId")
 @rxSqH
 def get_deploy_templates():
  return{"create":{"function":"create_authorizer"}}
class ApiGatewayV2Api(GenericBaseModel):
 @rxSqH
 def cloudformation_type():
  return "AWS::ApiGatewayV2::Api"
 def fetch_state(self,stack_name,resources):
  props=self.props
  api_name=(props["Body"].get("info",{}).get("title")if props.get("Body")else props["Name"])
  return self.fetch_details(api_name)
 @rxSqA
 def fetch_details(cls,api_name):
  client=aws_stack.connect_to_service("apigatewayv2")
  apis=client.get_apis()["Items"]
  result=([a for a in apis if a["Name"]==api_name]or[rxSqW])[0]
  return result
 def get_physical_resource_id(self,attribute,**kwargs):
  if attribute in REF_ID_ATTRS:
   return self.props.get("ApiId")
 @rxSqH
 def get_deploy_templates():
  def create_resource(resource_id,resources,resource_type,func,stack_name,*args):
   resource=resources[resource_id]
   resource_props=resource["Properties"]
   client=aws_stack.connect_to_service("apigatewayv2")
   body=resource_props.get("Body")
   if body:
    base_path=resource_props.get("Basepath")
    body.setdefault("info",{}).setdefault("title","api-%s"%short_uid())
    body=json.dumps(body)
    kwargs={"Basepath":base_path}if base_path else{}
    return client.import_api(Body=body,**kwargs)
   params=select_attributes(resource_props,["ApiKeySelectionExpression","CorsConfiguration","CredentialsArn","Description","DisableSchemaValidation","Name","ProtocolType","RouteKey","RouteSelectionExpression","Tags","Target","Version"])
   return client.create_api(**params)
  return{"create":{"function":create_resource}}
class ApiGatewayV2IntegrationResponse(GenericBaseModel):
 @rxSqH
 def cloudformation_type():
  return "AWS::ApiGatewayV2::IntegrationResponse"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("apigatewayv2")
  props=self.props
  api_id=self.resolve_refs_recursively(stack_name,props["ApiId"],resources)
  int_id=self.resolve_refs_recursively(stack_name,props["IntegrationId"],resources)
  resp_key=self.resolve_refs_recursively(stack_name,props["IntegrationResponseKey"],resources)
  responses=client.get_integration_responses(ApiId=api_id,IntegrationId=int_id).get("Items",[])
  result=[r for r in responses if r["IntegrationResponseKey"]==resp_key]
  return(result or[rxSqW])[0]
 def get_physical_resource_id(self,attribute=rxSqW,**kwargs):
  return self.props.get("IntegrationResponseId")
 def get_cfn_attribute(self,attribute_name):
  if attribute_name in REF_ID_ATTRS:
   return self.get_physical_resource_id(attribute_name)
  return rxSqB(ApiGatewayV2IntegrationResponse,self).get_cfn_attribute(attribute_name)
 @rxSqH
 def get_deploy_templates():
  return{"create":{"function":"create_integration_response"}}
class ApiGatewayV2Integration(GenericBaseModel):
 @rxSqH
 def cloudformation_type():
  return "AWS::ApiGatewayV2::Integration"
 def get_physical_resource_id(self,attribute,**kwargs):
  if attribute in REF_ID_ATTRS:
   return self.props.get("IntegrationId")
 @rxSqA
 def fetch_details(cls,api_id,rxSqP,method,uri=rxSqW):
  client=aws_stack.connect_to_service("apigatewayv2")
  resp=client.get_integrations(ApiId=api_id)
  if "Items" not in resp:
   return rxSqW
  integrations=[r for r in resp["Items"]if rxSqP==r.get("IntegrationType")and uri in[rxSqW,r.get("IntegrationUri")]and method in[rxSqW,r.get("IntegrationMethod")]]
  return(integrations or[rxSqW])[0]
class ApiGatewayV2Deployment(GenericBaseModel):
 @rxSqH
 def cloudformation_type():
  return "AWS::ApiGatewayV2::Deployment"
 def get_physical_resource_id(self,attribute,**kwargs):
  if attribute in REF_ID_ATTRS:
   return self.props.get("DeploymentId")
 @rxSqA
 def fetch_details(cls,api_id,stage_name,description):
  client=aws_stack.connect_to_service("apigatewayv2")
  apis=client.get_deployments(ApiId=api_id)["Items"]
  apis=[a for a in apis if a.get("StageName")==stage_name or a.get("Description")==description]
  return(apis or[rxSqW])[0]
class ApiGatewayV2Stage(GenericBaseModel):
 @rxSqH
 def cloudformation_type():
  return "AWS::ApiGatewayV2::Stage"
 @rxSqA
 def fetch_details(cls,api_id,stage_name):
  client=aws_stack.connect_to_service("apigatewayv2")
  return client.get_stage(ApiId=api_id,StageName=stage_name)or rxSqW
 def get_physical_resource_id(self,attribute,**kwargs):
  if attribute in REF_ID_ATTRS:
   return self.props.get("StageName")
class ApiGatewayV2RouteResponse(GenericBaseModel):
 @rxSqH
 def cloudformation_type():
  return "AWS::ApiGatewayV2::RouteResponse"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("apigatewayv2")
  props=self.props
  api_id=self.resolve_refs_recursively(stack_name,props["ApiId"],resources)
  route_id=self.resolve_refs_recursively(stack_name,props["RouteId"],resources)
  response_key=self.resolve_refs_recursively(stack_name,props["RouteResponseKey"],resources)
  responses=client.get_route_responses(ApiId=api_id,RouteId=route_id).get("Items",[])
  result=[r for r in responses if r["RouteResponseKey"]==response_key]
  return(result or[rxSqW])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("RouteResponseId")
 @rxSqH
 def get_deploy_templates():
  return{"create":{"function":"create_route_response"},"delete":{"function":"delete_route_response","parameters":["ApiId","RouteId","RouteResponseId"]}}
class ApiGatewayV2Route(GenericBaseModel):
 @rxSqH
 def cloudformation_type():
  return "AWS::ApiGatewayV2::Route"
 @rxSqA
 def fetch_details(cls,api_id,route_key):
  client=aws_stack.connect_to_service("apigatewayv2")
  routes=client.get_routes(ApiId=api_id).get("Items",[])
  result=[r for r in routes if r["RouteKey"]==route_key]
  return(result or[rxSqW])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("RouteId")
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.utils.aws import aws_stack
wRxlE=staticmethod
wRxlF=None
wRxlS=classmethod
from localstack.utils.common import select_attributes
from localstack_ext.services.cloudformation.models.servicediscovery import ServiceDiscoveryNamespace
class ApplicationAutoScalingPolicy(ServiceDiscoveryNamespace):
 @wRxlE
 def cloudformation_type():
  return "AWS::ApplicationAutoScaling::ScalingPolicy"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("application-autoscaling")
  props=self.props
  service_ns=props.get("ServiceNamespace")
  target_id=props.get("ScalingTargetId")
  if target_id:
   service_ns=target_id.split("|")[-1]
  pol_name=self.resolve_refs_recursively(stack_name,props["PolicyName"],resources)
  service_ns=self.resolve_refs_recursively(stack_name,service_ns,resources)
  policies=client.describe_scaling_policies(ServiceNamespace=service_ns)["ScalingPolicies"]
  policy=[p for p in policies if p["PolicyName"]==pol_name]
  return(policy or[wRxlF])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("PolicyARN")
 @wRxlS
 def get_deploy_templates(cls):
  def get_scaling_policy_params(params,**kwargs):
   attrs=("PolicyName","ServiceNamespace","ResourceId","ScalableDimension","PolicyType","StepScalingPolicyConfiguration","TargetTrackingScalingPolicyConfiguration")
   result=select_attributes(params,attrs)
   target_id=params.get("ScalingTargetId")
   if target_id:
    parts=target_id.split("|")
    result.setdefault("ResourceId",parts[0])
    result.setdefault("ScalableDimension",parts[1])
    result.setdefault("ServiceNamespace",parts[2])
   return result
  return{"create":{"function":"put_scaling_policy","parameters":get_scaling_policy_params},"delete":{"function":"delete_scaling_policy","parameters":["PolicyName","ServiceNamespace","ResourceId","ScalableDimension"]}}
class ApplicationAutoScalingScalableTarget(ServiceDiscoveryNamespace):
 @wRxlE
 def cloudformation_type():
  return "AWS::ApplicationAutoScaling::ScalableTarget"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("application-autoscaling")
  props=self.props
  resource_id=self.resolve_refs_recursively(stack_name,props["ResourceId"],resources)
  scal_dim=self.resolve_refs_recursively(stack_name,props["ScalableDimension"],resources)
  service_ns=self.resolve_refs_recursively(stack_name,props["ServiceNamespace"],resources)
  services=client.describe_scalable_targets(ServiceNamespace=service_ns)["ScalableTargets"]
  service=[s for s in services if s["ResourceId"]==resource_id and s["ScalableDimension"]==scal_dim and s["ServiceNamespace"]==service_ns]
  return(service or[wRxlF])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  if not self.state:
   return
  props=self.props
  return "service/%s|%s|%s"%(props["ResourceId"],props["ScalableDimension"],props["ServiceNamespace"])
 @wRxlS
 def get_deploy_templates(cls):
  return{"create":{"function":"register_scalable_target","parameters":["ServiceNamespace","ResourceId","ScalableDimension","MinCapacity","MaxCapacity","RoleARN","SuspendedState"]},"delete":{"function":"deregister_scalable_target","parameters":["ServiceNamespace","ResourceId","ScalableDimension"]}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from datetime import datetime
vyQwb=str
vyQwu=int
vyQwG=super
vyQwq=False
vyQwY=isinstance
vyQwg=hash
vyQwe=bool
vyQwX=True
vyQwx=list
vyQwB=map
vyQwM=None
from typing import Set
from localstack_ext.bootstrap.cpvcs.constants import(COMMIT_TXT_LAYOUT,REV_TXT_LAYOUT,STATE_TXT_LAYOUT,STATE_TXT_METADATA,VER_TXT_LAYOUT)
class CPVCSObj:
 def __init__(self,hash_ref:vyQwb):
  self.hash_ref:vyQwb=hash_ref
class StateFileRef(CPVCSObj):
 txt_layout=STATE_TXT_LAYOUT
 metadata_layout=STATE_TXT_METADATA
 def __init__(self,hash_ref:vyQwb,rel_path:vyQwb,file_name:vyQwb,size:vyQwu,service:vyQwb,region:vyQwb):
  vyQwG(StateFileRef,self).__init__(hash_ref)
  self.rel_path:vyQwb=rel_path
  self.file_name:vyQwb=file_name
  self.size:vyQwu=size
  self.service:vyQwb=service
  self.region:vyQwb=region
 def __str__(self):
  return self.txt_layout.format(size=self.size,service=self.service,region=self.region,hash_ref=self.hash_ref,file_name=self.file_name,rel_path=self.rel_path)
 def __eq__(self,other):
  if not other:
   return vyQwq
  if not vyQwY(other,StateFileRef):
   return vyQwq
  return(self.hash_ref==other.hash_ref and self.region==other.region and self.service==self.service and self.file_name==other.file_name and self.size==other.size)
 def __hash__(self):
  return vyQwg((self.hash_ref,self.region,self.service,self.file_name,self.size))
 def congruent(self,other)->vyQwe:
  if not other:
   return vyQwq
  if not vyQwY(other,StateFileRef):
   return vyQwq
  return(self.region==other.region and self.service==other.service and self.file_name==other.file_name and self.rel_path==other.rel_path)
 def any_congruence(self,others)->vyQwe:
  for other in others:
   if self.congruent(other):
    return vyQwX
  return vyQwq
 def metadata(self)->vyQwb:
  return self.metadata_layout.format(size=self.size,service=self.service,region=self.region)
class CPVCSNode(CPVCSObj):
 def __init__(self,hash_ref:vyQwb,state_files:Set[StateFileRef],parent_ptr:vyQwb):
  vyQwG(CPVCSNode,self).__init__(hash_ref)
  self.state_files:Set[StateFileRef]=state_files
  self.parent_ptr:vyQwb=parent_ptr
 def state_files_info(self)->vyQwb:
  return "\n".join(vyQwx(vyQwB(lambda state_file:vyQwb(state_file),self.state_files)))
class Commit:
 txt_layout=COMMIT_TXT_LAYOUT
 def __init__(self,tail_ptr:vyQwb,head_ptr:vyQwb,message:vyQwb,timestamp:vyQwb=vyQwb(datetime.now().timestamp()),delta_log_ptr:vyQwb=vyQwM):
  self.tail_ptr:vyQwb=tail_ptr
  self.head_ptr:vyQwb=head_ptr
  self.message:vyQwb=message
  self.timestamp:vyQwb=timestamp
  self.delta_log_ptr:vyQwb=delta_log_ptr
 def __str__(self):
  return self.txt_layout.format(tail_ptr=self.tail_ptr,head_ptr=self.head_ptr,message=self.message,timestamp=self.timestamp,log_hash=self.delta_log_ptr)
 def info_str(self,from_node:vyQwb,to_node:vyQwb)->vyQwb:
  return f"from: {from_node}, to: {to_node}, message: {self.message}, time: {datetime.fromtimestamp(float(self.timestamp))}"
class Revision(CPVCSNode):
 txt_layout=REV_TXT_LAYOUT
 def __init__(self,hash_ref:vyQwb,state_files:Set[StateFileRef],parent_ptr:vyQwb,creator:vyQwb,rid:vyQwb,revision_number:vyQwu,assoc_commit:Commit=vyQwM):
  vyQwG(Revision,self).__init__(hash_ref,state_files,parent_ptr)
  self.creator:vyQwb=creator
  self.rid:vyQwb=rid
  self.revision_number:vyQwu=revision_number
  self.assoc_commit=assoc_commit
 def __str__(self):
  return self.txt_layout.format(hash_ref=self.hash_ref,parent=self.parent_ptr,creator=self.creator,rid=self.rid,rev_no=self.revision_number,state_files=";".join(vyQwB(lambda state_file:vyQwb(state_file),self.state_files))if self.state_files else "",assoc_commit=self.assoc_commit)
class Version(CPVCSNode):
 txt_layout=VER_TXT_LAYOUT
 def __init__(self,hash_ref:vyQwb,state_files:Set[StateFileRef],parent_ptr:vyQwb,creator:vyQwb,comment:vyQwb,active_revision_ptr:vyQwb,outgoing_revision_ptrs:Set[vyQwb],incoming_revision_ptr:vyQwb,version_number:vyQwu):
  vyQwG(Version,self).__init__(hash_ref,state_files,parent_ptr)
  self.creator=creator
  self.comment=comment
  self.active_revision_ptr=active_revision_ptr
  self.outgoing_revision_ptrs=outgoing_revision_ptrs
  self.incoming_revision_ptr=incoming_revision_ptr
  self.version_number=version_number
 def __str__(self):
  return VER_TXT_LAYOUT.format(hash_ref=self.hash_ref,parent=self.parent_ptr,creator=self.creator,comment=self.comment,version_number=self.version_number,active_revision=self.active_revision_ptr,outgoing_revisions=";".join(self.outgoing_revision_ptrs),incoming_revision=self.incoming_revision_ptr,state_files=";".join(vyQwB(lambda stat_file:vyQwb(stat_file),self.state_files))if self.state_files else "")
 def info_str(self):
  return f"{self.version_number}, {self.creator}, {self.comment}"
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import REF_ID_ATTRS,GenericBaseModel
ThQqo=staticmethod
ThQqX=None
ThQqP=super
ThQqt=classmethod
ThQqj=isinstance
ThQqa=list
ThQqI=set
ThQqz=False
ThQqp=str
ThQqe=dict
ThQqg=len
from localstack.utils.aws import aws_stack
from localstack.utils.common import select_attributes,short_uid,to_bytes
from localstack_ext.utils.aws import aws_utils
class CloudFrontOriginAccessIdentity(GenericBaseModel):
 @ThQqo
 def cloudformation_type():
  return "AWS::CloudFront::CloudFrontOriginAccessIdentity"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("cloudfront")
  result=client.list_cloud_front_origin_access_identities()
  result=result.get("CloudFrontOriginAccessIdentityList",{}).get("Items",[])
  config=self.props.get("CloudFrontOriginAccessIdentityConfig",{})
  result=[r for r in result if r["Comment"]==config.get("Comment")]
  return result and result[0]or ThQqX
 @ThQqo
 def get_deploy_templates():
  def _params(params,**kwargs):
   config=params.get("CloudFrontOriginAccessIdentityConfig")
   config["CallerReference"]=short_uid()
   return params
  return{"create":{"function":"create_cloud_front_origin_access_identity","parameters":_params}}
class CloudFrontFunction(GenericBaseModel):
 @ThQqo
 def cloudformation_type():
  return "AWS::CloudFront::Function"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("cloudfront")
  result=client.list_functions()
  result=result.get("FunctionList",{}).get("Items",[])
  props=self.props
  result=[r for r in result if r["Name"]==props.get("Name")]
  return result and result[0]or ThQqX
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("Name")
 def get_cfn_attribute(self,attribute_name):
  if attribute_name=="FunctionARN":
   return aws_utils.get_cloudfront_function_arn(self.props.get("Name"))
  return ThQqP(CloudFrontFunction,self).get_cfn_attribute(attribute_name)
 @ThQqt
 def get_deploy_templates(cls):
  def _params(params,**kwargs):
   result=select_attributes(params,["Name","FunctionCode","FunctionConfig"])
   if result.get("FunctionCode"):
    result["FunctionCode"]=to_bytes(result["FunctionCode"])
   return result
  def _delete_function(resource_id,resources,resource_type,func,stack_name,*args):
   resource=cls(resources[resource_id])
   client=aws_stack.connect_to_service("cloudfront")
   name=resource.props.get("Name")
   etag=client.describe_function(Name=name).get("ETag")
   return client.delete_function(Name=name,IfMatch=etag)
  return{"create":{"function":"create_function","parameters":_params},"delete":{"function":_delete_function}}
class CloudFrontOriginRequestPolicy(GenericBaseModel):
 @ThQqo
 def cloudformation_type():
  return "AWS::CloudFront::OriginRequestPolicy"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("cloudfront")
  result=client.list_origin_request_policies()
  result=result.get("OriginRequestPolicyList",{}).get("Items",[])
  result=[r.get("OriginRequestPolicy",{})for r in result]
  result=[{"Id":r["Id"],**r.get("OriginRequestPolicyConfig",{})}for r in result]
  policy_name=self.props.get("OriginRequestPolicyConfig",{}).get("Name")
  policy_name=self.resolve_refs_recursively(stack_name,policy_name,resources)
  result=[r for r in result if r.get("Name")==policy_name]
  return result and result[0]or ThQqX
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("Id")
 @ThQqt
 def get_deploy_templates(cls):
  def _delete_policy(resource_id,resources,resource_type,func,stack_name,*args):
   client=aws_stack.connect_to_service("cloudfront")
   resource=cls(resources[resource_id])
   props=resource.props
   policy_id=props.get("OriginRequestPolicy").get("Id")
   etag=client.get_origin_request_policy(Id=policy_id).get("ETag")
   return client.delete_origin_request_policy(Id=policy_id,IfMatch=etag)
  return{"create":{"function":"create_origin_request_policy"},"delete":{"function":_delete_policy}}
class CloudFrontDistribution(GenericBaseModel):
 @ThQqo
 def cloudformation_type():
  return "AWS::CloudFront::Distribution"
 def get_physical_resource_id(self,attribute,**kwargs):
  distr_id=self.props.get("Id")
  if attribute in REF_ID_ATTRS:
   return distr_id
  return aws_utils.get_cloudfront_distribution_arn(distr_id)
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("cloudfront")
  distrs=client.list_distributions()["DistributionList"].get("Items",[])
  props=self.props["DistributionConfig"]
  origins=props.get("Origins",[])
  origins=origins if ThQqj(origins,ThQqa)else origins.get("Items")
  origins_to_create=ThQqI([orig["DomainName"]for orig in origins])
  aliases=props.get("Aliases",[])
  aliases=aliases if ThQqj(aliases,ThQqa)else aliases.get("Items")
  aliases_to_create=ThQqI(aliases)
  for distr in distrs:
   origins=ThQqI([orig.get("DomainName")for orig in distr.get("Origins",{})["Items"]])
   aliases=ThQqI(distr.get("Aliases",{}).get("Items",[]))
   if origins_to_create==origins and aliases_to_create==aliases:
    return distr
 @ThQqo
 def get_deploy_templates():
  def lambda_get_distribution_config(params,**kwargs):
   config=params["DistributionConfig"]
   config["CallerReference"]=config.get("CallerReference")or short_uid()
   config["Comment"]=config.get("Comment")or ""
   config["IsIPV6Enabled"]=config.pop("IPV6Enabled",ThQqz)
   config["Enabled"]=ThQqp(config.get("Enabled")).lower()=="true"
   cert=config["ViewerCertificate"]=config.get("ViewerCertificate",{})
   custom_errors=config.get("CustomErrorResponses",[])
   custom_errors=(custom_errors.get("Items",[])if ThQqj(custom_errors,ThQqe)else custom_errors)
   for item in custom_errors:
    item["ResponseCode"]=ThQqp(item.get("ResponseCode"))
   cache=config["DefaultCacheBehavior"]
   cache["TrustedSigners"]=cache.get("TrustedSigners")or{"Quantity":0,"Enabled":ThQqz}
   cache["MinTTL"]=cache.get("MinTTL")or 600
   forwarded=cache["ForwardedValues"]=cache.get("ForwardedValues",{})
   forwarded["QueryString"]=ThQqp(forwarded.get("QueryString")).lower()=="true"
   if "AcmCertificateArn" in cert:
    cert["ACMCertificateArn"]=cert.pop("AcmCertificateArn")
   if "SslSupportMethod" in cert:
    cert["SSLSupportMethod"]=cert.pop("SslSupportMethod")
   if "IamCertificateId" in cert:
    cert["IAMCertificateId"]=cert.pop("IamCertificateId")
   params["Tags"]={"Items":params.get("Tags")or[]}
   def convert_list(key,parent=ThQqX):
    parent=config if parent is ThQqX else parent
    items=parent.get(key)or[]
    if ThQqj(items,ThQqe)and "Items" in items and "Quantity" in items:
     return
    parent[key]={"Quantity":ThQqg(items),"Items":items}
   cache_behavs=config.get("CacheBehaviors",[])
   cache_behavs=(cache_behavs.get("Items",[])if ThQqj(cache_behavs,ThQqe)else cache_behavs)
   for cb in cache_behavs:
    convert_list("AllowedMethods",cb)
    cb["AllowedMethods"]["CachedMethods"]=cb.pop("CachedMethods",[])
    convert_list("CachedMethods",cb["AllowedMethods"])
   for origin in config["Origins"]:
    custom_origin_config=origin.get("CustomOriginConfig")
    if not custom_origin_config:
     continue
    custom_origin_config["OriginSslProtocols"]=custom_origin_config.pop("OriginSSLProtocols",[])
    custom_origin_config["HTTPPort"]=custom_origin_config.get("HTTPPort",80)
    custom_origin_config["HTTPSPort"]=custom_origin_config.get("HTTPSPort",443)
    convert_list("OriginSslProtocols",custom_origin_config)
   convert_list("Aliases")
   convert_list("Origins")
   convert_list("OriginGroups")
   convert_list("CustomErrorResponses")
   convert_list("CacheBehaviors")
   convert_list("AllowedMethods",cache)
   cache_behav=config.get("DefaultCacheBehavior",{})
   cache_behav["AllowedMethods"]=cache_behav.get("AllowedMethods")
   cache_behav["AllowedMethods"]["CachedMethods"]=config.get("DefaultCacheBehavior",{}).pop("CachedMethods",[])
   convert_list("CachedMethods",cache_behav["AllowedMethods"])
   if cache_behav.get("ForwardedValues")and not cache_behav["ForwardedValues"].get("Cookies"):
    cache_behav["ForwardedValues"]["Cookies"]={"Forward":"All"}
   restrs=config.get("Restrictions",{})
   if restrs.get("GeoRestriction")and "Quantity" not in restrs["GeoRestriction"]:
    restrs["GeoRestriction"]["Quantity"]=0
    restrs["GeoRestriction"]["Items"]=[]
   return params
  return{"create":{"function":"create_distribution_with_tags","parameters":{"DistributionConfigWithTags":lambda_get_distribution_config}}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import getpass
QWxVI=object
QWxVh=staticmethod
QWxVz=False
QWxVM=Exception
QWxVj=None
QWxVF=input
QWxVr=list
import json
import logging
import os
import sys
from localstack.config import dirs
from localstack.constants import API_ENDPOINT
from localstack.utils.common import FileMappedDocument,call_safe,safe_requests,to_str
LOG=logging.getLogger(__name__)
AUTH_CACHE_FILE="auth.json"
class AuthProvider(QWxVI):
 @QWxVh
 def name():
  raise
 def get_or_create_token(self,username,password,headers):
  pass
 def get_user_for_token(self,token):
  pass
 @QWxVh
 def providers():
  return{c.name():c for c in AuthProvider.__subclasses__()}
 @QWxVh
 def get(provider,raise_error=QWxVz):
  provider_class=AuthProvider.providers().get(provider)
  if not provider_class:
   msg='Unable to find auth provider class "%s"'%provider
   LOG.warning(msg)
   if raise_error:
    raise QWxVM(msg)
   return QWxVj
  return provider_class()
class AuthProviderInternal(AuthProvider):
 @QWxVh
 def name():
  return "internal"
 def get_or_create_token(self,username,password,headers):
  data={"username":username,"password":password}
  response=safe_requests.post("%s/user/signin"%API_ENDPOINT,json.dumps(data),headers=headers)
  if response.status_code>=400:
   return
  try:
   result=json.loads(to_str(response.content or "{}"))
   return result["token"]
  except QWxVM:
   pass
 def read_credentials(self,username):
  print("Please provide your login credentials below")
  if not username:
   sys.stdout.write("Username: ")
   sys.stdout.flush()
   username=QWxVF()
  password=getpass.getpass()
  return username,password,{}
 def get_user_for_token(self,token):
  raise QWxVM("Not implemented")
def get_auth_cache()->FileMappedDocument:
 return FileMappedDocument(os.path.join(dirs.cache,AUTH_CACHE_FILE),mode=0o600)
def login(provider,username=QWxVj):
 auth_provider=AuthProvider.get(provider)
 if not auth_provider:
  providers=QWxVr(AuthProvider.providers().keys())
  raise QWxVM('Unknown provider "%s", should be one of %s'%(provider,providers))
 username,password,headers=auth_provider.read_credentials(username)
 print("Verifying credentials ... (this may take a few moments)")
 token=auth_provider.get_or_create_token(username,password,headers)
 if not token:
  raise QWxVM("Unable to verify login credentials - please try again")
 cache=get_auth_cache()
 cache.update({"provider":provider,"username":username,"token":token})
 call_safe(cache.save,exception_message="error saving authentication information")
def logout():
 cache=get_auth_cache()
 cache.clear()
 cache.save()
def json_loads(s):
 return json.loads(to_str(s))
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import logging
CWwjA=str
CWwjx=isinstance
CWwjy=False
CWwjI=Exception
CWwjM=True
CWwjN=bool
import os
import threading
import traceback
from typing import List,Union
import localstack
from localstack import constants
from localstack.services.install import Installer,InstallerRepository
from localstack.utils.common import(download,in_docker,is_command_available,is_debian,now,rm_rf,run)
from localstack_ext.bootstrap.licensing import api_key_configured,prepare_environment
from localstack_ext.constants import ARTIFACTS_REPO
LOG=logging.getLogger(__name__)
RULE_ENGINE_INSTALL_URL="https://github.com/whummer/serverless-iot-offline"
H2_DOWNLOAD_URL="http://www.h2database.com/h2-2019-10-14.zip"
SSL_CERT_URL="%s/raw/master/local-certs/server.key"%ARTIFACTS_REPO
SSL_CERT_URL_FALLBACK="{api_endpoint}/proxy/localstack.cert.key"
INFRA_DIR=os.path.join(os.path.dirname(localstack.__file__),"infra")
LOCALSTACK_DIR=os.path.dirname(localstack.__file__)
POSTGRES_LIB_FOLDER="/usr/lib/postgresql/11/lib"
INSTALL_LOCK=threading.RLock()
def install_libs():
 install_iot_rule_engine()
 install_postgres()
 install_timescaledb()
 install_redis()
 install_mqtt()
def install_iot_rule_engine():
 target_dir=LOCALSTACK_DIR
 main_file=os.path.join(target_dir,"node_modules","serverless-iot-offline","query.js")
 if not os.path.exists(main_file):
  LOG.info("Installing IoT rule engine. This may take a while.")
  run("cd %s; npm install %s"%(target_dir,RULE_ENGINE_INSTALL_URL))
 return main_file
def install_postgres():
 if not in_docker():
  return
 check_or_install("psql","postgresql-11","RDS")
 check_or_install("pg_config",["postgresql-server-dev-11","libpq-dev"],"RDS")
 if not is_debian():
  return
 plpython_lib=f"{POSTGRES_LIB_FOLDER}/plpython3.so"
 if os.path.exists(plpython_lib):
  return
 install_package("postgresql-plpython3-11","RDS")
def install_timescaledb():
 if not in_docker():
  return
 if os.path.exists(f"{POSTGRES_LIB_FOLDER}/timescaledb.so"):
  return
 check_or_install("gcc",["cmake","gcc","git"],"Timestream")
 ts_dir="/tmp/timescaledb"
 tag="2.0.0-rc4"
 run("cd /tmp; git clone https://github.com/timescale/timescaledb.git")
 run("cd %s; git checkout %s; ./bootstrap -DREGRESS_CHECKS=OFF; cd build; make; make install"%(ts_dir,tag))
 rm_rf("/tmp/timescaledb")
def install_redis():
 check_or_install("redis-server","redis-server","ElastiCache")
 return "redis-server"
def install_mqtt():
 check_or_install("mosquitto","mosquitto","IoT")
 return "mosquitto"
def install_package(packages:Union[List[CWwjA],CWwjA],api_name:CWwjA):
 if not(is_debian()and in_docker()):
  LOG.warning("Unable to install dependencies for %s API. Please install packages %s (or equivalent) on your system manually!",api_name,packages)
  return
 if CWwjx(packages,CWwjA):
  packages=packages.split()
 LOG.info("Downloading dependencies for %s API. This may take a while."%api_name)
 with INSTALL_LOCK:
  run(["apt-get","update"],shell=CWwjy)
  run(["apt-get","install","-y","--no-install-recommends"]+packages,shell=CWwjy)
def check_or_install(command:CWwjA,packages:Union[List[CWwjA],CWwjA],api:CWwjA):
 if not is_command_available(command):
  install_package(packages,api)
def setup_ssl_cert():
 from localstack.services import generic_proxy
 target_file=generic_proxy.get_cert_pem_file_path()
 cache_duration_secs=6*60*60
 if os.path.exists(target_file):
  mod_time=os.path.getmtime(target_file)
  if mod_time>(now()-cache_duration_secs):
   return
 try:
  return download_github_artifact(SSL_CERT_URL,target_file)
 except CWwjI:
  url=SSL_CERT_URL_FALLBACK.format(api_endpoint=constants.API_ENDPOINT)
  try:
   return download(url,target_file)
  except CWwjI:
   LOG.info("Unable to download local test SSL certificate from %s to %s",url,target_file)
   raise
def download_github_artifact(url,target_file):
 def do_download(url,print_error=CWwjy):
  try:
   download(url,target_file)
   return CWwjM
  except CWwjI as e:
   if print_error:
    LOG.info("Unable to download Github artifact from from %s to %s: %s %s"%(url,target_file,e,traceback.format_exc()))
 result=do_download(url)
 if not result:
  url=url.replace("https://github.com","https://cdn.jsdelivr.net/gh")
  url=url.replace("/raw/master/","@master/")
  do_download(url,CWwjM)
def install_azure():
 from localstack_ext.services.azure import api_specs
 api_specs.download_api_specs()
class ExtInstallerRepository(InstallerRepository):
 name="ext"
 def should_load(self)->CWwjN:
  return api_key_configured()
 def load(self,*args,**kwargs):
  LOG.debug("Preparing Pro environment for LocalStack Package Manager.")
  with prepare_environment():
   LOG.debug("Pro environment has successfully been prepared.")
 def get_installer(self)->List[Installer]:
  return[("iot-rule-engine",install_iot_rule_engine),("postgres",install_postgres),("timescaledb",install_timescaledb),("redis",install_redis),("mqtt",install_mqtt),("azure",install_azure)]
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import REF_ID_ATTRS,GenericBaseModel
HSdMO=str
HSdMG=staticmethod
HSdMu=None
HSdMe=super
from localstack.utils.aws import aws_stack
from localstack.utils.common import select_attributes,short_uid
def canonicalize_name(name:HSdMO)->HSdMO:
 if name[-1]!=".":
  return f"{name}."
 return name
class Route53HostedZone(GenericBaseModel):
 @HSdMG
 def cloudformation_type():
  return "AWS::Route53::HostedZone"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("route53")
  result=client.list_hosted_zones()["HostedZones"]
  canonical_name=canonicalize_name(self.props.get("Name"))
  result=[z for z in result if z["Name"]==canonical_name]
  return(result or[HSdMu])[0]
 def get_physical_resource_id(self,attribute=HSdMu,**kwargs):
  return self.props.get("Id")
 def get_cfn_attribute(self,attribute_name):
  if attribute_name=="NameServers":
   return[self.props.get("Name")]
  if attribute_name in REF_ID_ATTRS:
   return self.get_physical_resource_id()
  return HSdMe(Route53HostedZone,self).get_cfn_attribute(attribute_name)
 @HSdMG
 def get_deploy_templates():
  def _get_params(resource_props,resources,resource_id,*args,**kwargs):
   resource=Route53HostedZone(resources[resource_id])
   props=resource.props
   result=select_attributes(props,["HostedZoneConfig","Name"])
   vpcs=props.get("VPCs",[])
   result["VPC"]=vpcs and vpcs[0]or{}
   result["CallerReference"]=short_uid()
   return result
  def _get_id(resource_props,resources,resource_id,*args,**kwargs):
   resource=Route53HostedZone(resources[resource_id])
   return resource.props.get("Id")
  return{"create":{"function":"create_hosted_zone","parameters":_get_params},"delete":{"function":"delete_hosted_zone","parameters":{"Id":_get_id}}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import GenericBaseModel
jlkgD=staticmethod
jlkgI=None
jlkgV=super
jlkgs=classmethod
jlkgu=int
from localstack.utils.aws import aws_stack
class RDSDBSubnetGroup(GenericBaseModel):
 @jlkgD
 def cloudformation_type():
  return "AWS::RDS::DBSubnetGroup"
 def get_physical_resource_id(self,attribute=jlkgI,**kwargs):
  return self.props.get("DBSubnetGroupName")
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("rds")
  group_name=self.resolve_refs_recursively(stack_name,self.props.get("DBSubnetGroupName"),resources)
  group=client.describe_db_subnet_groups()["DBSubnetGroups"]
  match=[i for i in group if i["DBSubnetGroupName"]==group_name]
  return(match or[jlkgI])[0]
 @jlkgD
 def get_deploy_templates():
  return{"create":{"function":"create_db_subnet_group"},"delete":{"function":"delete_db_subnet_group","parameters":{"DBSubnetGroupName":"DBSubnetGroupName"}}}
class RDSDBCluster(GenericBaseModel):
 @jlkgD
 def cloudformation_type():
  return "AWS::RDS::DBCluster"
 def get_physical_resource_id(self,attribute=jlkgI,**kwargs):
  return self.props.get("DBClusterIdentifier")
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("rds")
  clusters=client.describe_db_clusters().get("DBClusters",[])
  cluster_id=self.resolve_refs_recursively(stack_name,self.props.get("DBClusterIdentifier"),resources)
  match=[c for c in clusters if c["DBClusterIdentifier"]==cluster_id]
  return(match or[jlkgI])[0]
 def get_cfn_attribute(self,attribute):
  if attribute=="Endpoint.Address":
   return "localhost"
  props=self.props
  if attribute=="Endpoint.Port":
   return props.get("Port")or props.get("Endpoint",{}).get("Port")
  return jlkgV(RDSDBCluster,self).get_cfn_attribute(attribute)
 @jlkgD
 def get_deploy_templates():
  attrs=["AvailabilityZones","BackupRetentionPeriod","CharacterSetName","DatabaseName","DBClusterIdentifier","DBClusterParameterGroupName","VpcSecurityGroupIds","DBSubnetGroupName","Engine","EngineVersion","Port","MasterUsername","MasterUserPassword","OptionGroupName","PreferredBackupWindow","PreferredMaintenanceWindow","ReplicationSourceIdentifier","Tags","StorageEncrypted","KmsKeyId","PreSignedUrl","EnableIAMDatabaseAuthentication","BacktrackWindow","EnableCloudwatchLogsExports","EngineMode","ScalingConfiguration","DeletionProtection","GlobalClusterIdentifier","EnableHttpEndpoint","CopyTagsToSnapshot","Domain","DomainIAMRoleName","EnableGlobalWriteForwarding","SourceRegion"]
  def _params(params,**kwargs):
   params={k:v for k,v in params.items()if k in attrs}
   return params
  result={"create":{"function":"create_db_cluster","parameters":_params}}
  return result
class RDSDBParameterGroup(GenericBaseModel):
 @jlkgD
 def cloudformation_type():
  return "AWS::RDS::DBParameterGroup"
 def get_physical_resource_id(self,attribute=jlkgI,**kwargs):
  return self.props.get("DBParameterGroupName")
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("rds")
  props=self.props
  descr=self.resolve_refs_recursively(stack_name,props.get("Description"),resources)
  family=self.resolve_refs_recursively(stack_name,props.get("Family"),resources)
  groups=client.describe_db_parameter_groups()["DBParameterGroups"]
  match=[g for g in groups if g["Family"]==family and g["Description"]==descr]
  return(match or[jlkgI])[0]
 @jlkgD
 def get_deploy_templates():
  return{"create":{"function":"create_db_parameter_group","parameters":{"DBParameterGroupName":"DBParameterGroupName","DBParameterGroupFamily":"Family","Description":"Description","Tags":"Tags"}},"delete":{"function":"delete_db_parameter_group","parameters":["DBParameterGroupName"]}}
class RDSDBInstance(GenericBaseModel):
 @jlkgD
 def cloudformation_type():
  return "AWS::RDS::DBInstance"
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("DBInstanceIdentifier")
 def get_cfn_attribute(self,attribute):
  if attribute=="Endpoint.Address":
   return "localhost"
  props=self.props
  if attribute=="Endpoint.Port":
   return props.get("Endpoint",{}).get("Port")or props.get("Port")
  return jlkgV(RDSDBInstance,self).get_cfn_attribute(attribute)
 @jlkgs
 def fetch_details(cls,db_name):
  client=aws_stack.connect_to_service("rds")
  instances=client.describe_db_instances()["DBInstances"]
  match=[i for i in instances if i["DBName"]==db_name]
  return(match or[jlkgI])[0]
 @jlkgD
 def get_deploy_templates():
  attrs=["DBName","DBInstanceIdentifier","AllocatedStorage","DBInstanceClass","Engine","MasterUsername","MasterUserPassword","DBSecurityGroups","AvailabilityZone","DBSubnetGroupName","PreferredMaintenanceWindow","DBParameterGroupName","BackupRetentionPeriod","PreferredBackupWindow","Port","MultiAZ","EngineVersion","AutoMinorVersionUpgrade","LicenseModel","Iops","OptionGroupName","CharacterSetName","NcharCharacterSetName","PubliclyAccessible","Tags","DBClusterIdentifier","StorageType","TdeCredentialArn","TdeCredentialPassword","StorageEncrypted","KmsKeyId","Domain","CopyTagsToSnapshot","MonitoringInterval","MonitoringRoleArn","DomainIAMRoleName","PromotionTier","Timezone","EnableIAMDatabaseAuthentication","EnablePerformanceInsights","PerformanceInsightsKMSKeyId","PerformanceInsightsRetentionPeriod","EnableCloudwatchLogsExports","ProcessorFeatures","DeletionProtection","MaxAllocatedStorage","EnableCustomerOwnedIp"]
  return{"create":{"function":"create_db_instance","parameters":attrs+[{"VpcSecurityGroupIds":"VPCSecurityGroups"}],"types":{"AllocatedStorage":jlkgu}},"delete":{"function":"delete_db_instance","parameters":["DBInstanceIdentifier"]}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import GenericBaseModel
znyDT=staticmethod
znyDh=None
znyDU=classmethod
from localstack.utils.aws import aws_stack
from localstack.utils.common import clone
class KafkaCluster(GenericBaseModel):
 @znyDT
 def cloudformation_type():
  return "AWS::MSK::Cluster"
 def get_physical_resource_id(self,attribute=znyDh,**kwargs):
  return self.props.get("ClusterArn")
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("kafka")
  cluster_name=self.resolve_refs_recursively(stack_name,self.props["ClusterName"],resources)
  clusters=client.list_clusters()["ClusterInfoList"]
  clusters=[c for c in clusters if c["ClusterName"]==cluster_name]
  return(clusters or[znyDh])[0]
 @znyDU
 def get_deploy_templates(cls):
  def _create_params(params,resource_id,resources,**kwargs):
   resource=cls(resources[resource_id])
   props=clone(resource.props)
   storage_info=props.get("BrokerNodeGroupInfo",{}).get("StorageInfo",{})
   if "EBSStorageInfo" in storage_info:
    storage_info["EbsStorageInfo"]=storage_info.pop("EBSStorageInfo")
   return props
  def _delete_params(params,resource_id,resources,**kwargs):
   resource=cls(resources[resource_id])
   return{"ClusterArn":resource.props.get("ClusterArn")}
  return{"create":{"function":"create_cluster","parameters":_create_params},"delete":{"function":"delete_cluster","parameters":_delete_params}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.deployment_utils import lambda_keys_to_lower
zMhVC=staticmethod
zMhVw=None
zMhVu=classmethod
from localstack.services.cloudformation.service_models import GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack_ext.services.cloudformation.service_models import lambda_rename_attributes
class IoTAnalyticsChannel(GenericBaseModel):
 @zMhVC
 def cloudformation_type():
  return "AWS::IoTAnalytics::Channel"
 def fetch_state(self,stack_name,resources):
  channel_name=self.resolve_refs_recursively(stack_name,self.props.get("ChannelName"),resources)
  client=aws_stack.connect_to_service("iotanalytics")
  channels=client.list_channels(maxResults=500)["channelSummaries"]
  match=[c for c in channels if c.get("channelName")==channel_name]
  return(match or[zMhVw])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("ChannelName")
 @zMhVu
 def get_deploy_templates(cls):
  return{"create":{"function":"create_channel","parameters":{"channelName":"ChannelName","channelStorage":"ChannelStorage","retentionPeriod":"RetentionPeriod","tags":lambda_keys_to_lower("Tags")}},"delete":{"function":"delete_channel","parameters":{"channelName":"ChannelName"}}}
class IoTAnalyticsDataset(GenericBaseModel):
 @zMhVC
 def cloudformation_type():
  return "AWS::IoTAnalytics::Dataset"
 def fetch_state(self,stack_name,resources):
  dataset_name=self.resolve_refs_recursively(stack_name,self.props.get("DatasetName"),resources)
  client=aws_stack.connect_to_service("iotanalytics")
  datasets=client.list_datasets(maxResults=500)["datasetSummaries"]
  match=[d for d in datasets if d.get("datasetName")==dataset_name]
  return(match or[zMhVw])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("DatasetName")
 @zMhVu
 def get_deploy_templates(cls):
  return{"create":{"function":"create_dataset","parameters":lambda_rename_attributes({"scheduleExpression":"expression"},func=lambda_keys_to_lower())},"delete":{"function":"delete_dataset","parameters":{"datasetName":"DatasetName"}}}
class IoTAnalyticsPipeline(GenericBaseModel):
 @zMhVC
 def cloudformation_type():
  return "AWS::IoTAnalytics::Pipeline"
 def fetch_state(self,stack_name,resources):
  pipeline_name=self.resolve_refs_recursively(stack_name,self.props.get("PipelineName"),resources)
  client=aws_stack.connect_to_service("iotanalytics")
  pipelines=client.list_pipelines(maxResults=500)["pipelineSummaries"]
  match=[p for p in pipelines if p.get("pipelineName")==pipeline_name]
  return(match or[zMhVw])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("PipelineName")
 @zMhVu
 def get_deploy_templates(cls):
  return{"create":{"function":"create_pipeline","parameters":{"pipelineName":"PipelineName","pipelineActivities":lambda_keys_to_lower("PipelineActivities"),"tags":lambda_keys_to_lower("Tags")}},"delete":{"function":"delete_pipeline","parameters":{"pipelineName":"PipelineName"}}}
class IoTAnalyticsDatastore(GenericBaseModel):
 @zMhVC
 def cloudformation_type():
  return "AWS::IoTAnalytics::Datastore"
 def fetch_state(self,stack_name,resources):
  datastore_name=self.resolve_refs_recursively(stack_name,self.props.get("DatastoreName"),resources)
  client=aws_stack.connect_to_service("iotanalytics")
  datastores=client.list_datastores(maxResults=500)["datastoreSummaries"]
  match=[d for d in datastores if d.get("datastoreName")==datastore_name]
  return(match or[zMhVw])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("DatastoreName")
 @zMhVu
 def get_deploy_templates(cls):
  return{"create":{"function":"create_datastore","parameters":lambda_keys_to_lower()},"delete":{"function":"delete_datastore","parameters":{"datastoreName":"DatastoreName"}}}
# Created by pyminifier (https://github.com/liftoff/pyminifier)



from localstack.services.cloudformation.service_models import GenericBaseModel
hriGD=None
hriGE=classmethod
hriGR=staticmethod
from localstack.utils.aws import aws_stack
class ServiceDiscoveryNamespace(GenericBaseModel):
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("servicediscovery")
  ns_name=self.resolve_refs_recursively(stack_name,self.props["Name"],resources)
  namespaces=client.list_namespaces()["Namespaces"]
  namespace=[ns for ns in namespaces if ns["Type"]==self._type()and ns["Name"]==ns_name]
  return(namespace or[hriGD])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  return self.props.get("Id")
 @hriGE
 def get_deploy_templates(cls):
  def delete_params(resource_props,resources,resource_id,**kwargs):
   resource=cls(resources[resource_id])
   return{"Id":resource.props.get("Id")}
  create_func=cls._create_function()
  return{"create":{"function":create_func},"delete":{"function":"delete_namespace","parameters":delete_params}}
 def _type(self):
  raise
class ServiceDiscoveryService(ServiceDiscoveryNamespace):
 @hriGR
 def cloudformation_type():
  return "AWS::ServiceDiscovery::Service"
 def fetch_state(self,stack_name,resources):
  client=aws_stack.connect_to_service("servicediscovery")
  service_name=self.resolve_refs_recursively(stack_name,self.props["Name"],resources)
  services=client.list_services()["Services"]
  service=[s for s in services if s["Name"]==service_name]
  return(service or[hriGD])[0]
 def get_physical_resource_id(self,attribute,**kwargs):
  if attribute=="Arn":
   return self.props.get("Arn")
  return self.props.get("Id")
 @hriGE
 def get_deploy_templates(cls):
  def delete_params(resource_props,resources,resource_id,**kwargs):
   resource=cls(resources[resource_id])
   return{"Id":resource.props.get("Id")}
  return{"create":{"function":"create_service"},"delete":{"function":"delete_service","parameters":delete_params}}
class ServiceDiscoveryHttpNamespace(ServiceDiscoveryNamespace):
 @hriGR
 def cloudformation_type():
  return "AWS::ServiceDiscovery::HttpNamespace"
 @hriGE
 def _type(cls):
  return "HTTP"
 @hriGE
 def _create_function(cls):
  return "create_http_namespace"
class ServiceDiscoveryPublicDnsNamespace(ServiceDiscoveryNamespace):
 @hriGR
 def cloudformation_type():
  return "AWS::ServiceDiscovery::PublicDnsNamespace"
 @hriGE
 def _type(cls):
  return "DNS_PUBLIC"
 @hriGE
 def _create_function(cls):
  return "create_public_dns_namespace"
class ServiceDiscoveryPrivateDnsNamespace(ServiceDiscoveryNamespace):
 @hriGR
 def cloudformation_type():
  return "AWS::ServiceDiscovery::PrivateDnsNamespace"
 @hriGE
 def _type(cls):
  return "DNS_PRIVATE"
 @hriGE
 def _create_function(cls):
  return "create_private_dns_namespace"
# Created by pyminifier (https://github.com/liftoff/pyminifier)



NIL_PTR = "NIL"
CPVCS_DIR = ".cpvcs"
DEFAULT_POD_DIR = "cpvcs-pod"
OBJ_STORE_DIR = "objects"
REV_SUB_DIR = "rev"
VER_SUB_DIR = "ver"
DELTA_LOG_DIR = "deltas"
HEAD_FILE = "HEAD"
VER_LOG_FILE = "VER_LOG"
MAX_VER_FILE = "MAX_VER"
KNOWN_VER_FILE = "KNOWN_VER"
REMOTE_FILE = ".REMOTE"
REFS_DIR = "refs"
VER_SYMLINK = "{ver_no}"
REV_SYMLINK = "{rid}_{rev_no}"
META_ZIP = "version_{version_no}_meta"
STATE_ZIP = "version_{version_no}"
COMMIT_FILE = "commit_no_{commit_no}.json"
METAMODELS_FILE = "metamodels.json"
VERSION_SERVICE_INFO_FILE = "version_info.json"
COMPRESSION_FORMAT = "zip"
VERSION_SPACE_FILES = [KNOWN_VER_FILE, MAX_VER_FILE, VER_LOG_FILE]
VERSION_SPACE_DIRS = [OBJ_STORE_DIR, REFS_DIR]
VERSION_SPACE_ARCHIVE = "version_space.zip"
STATE_TXT_LAYOUT = "size:{size}, service:{service}, region:{region}, key:{hash_ref},file_name:{file_name}, rel_path:{rel_path}"
STATE_TXT_METADATA = "size: {size}, service:{service}, region: {region}"
COMMIT_TXT_LAYOUT = (
    "tail:{tail_ptr}, head:{head_ptr}, message:{message}, timestamp:{timestamp}, log_key:{log_hash}"
)
VER_TXT_LAYOUT = """parent_ptr={parent}
hash_ref={hash_ref}
creator={creator}
comment={comment}
version_number={version_number}
active_revision_ptr={active_revision}
outgoing_revision_ptrs={outgoing_revisions}
incoming_revision_ptr={incoming_revision}
state_files={state_files}
"""
REV_TXT_LAYOUT = """parent_ptr={parent}
hash_ref={hash_ref}
creator={creator}
rid={rid}
revision_number={rev_no}
state_files={state_files}
assoc_commit={assoc_commit}
"""
VER_LOG_STRUCTURE = "{author};{ver_no};{rev_rid_no}"



import json
Lhtri=None
Lhtrp=object
LhtrK=Exception
LhtrY=bool
LhtrP=False
Lhtrf=int
Lhtrq=str
LhtrA=set
LhtrE=property
Lhtrd=classmethod
LhtrF=True
LhtrM=staticmethod
Lhtrj=open
LhtrC=map
LhtrV=range
Lhtrk=len
Lhtrx=getattr
Lhtrc=type
LhtrH=isinstance
LhtrT=list
import logging
import os
import re
import traceback
from typing import Dict,List,Set
from zipfile import ZipFile
import requests
import yaml
from dulwich import porcelain
from dulwich.client import get_transport_and_path_from_url
from dulwich.repo import Repo
from localstack import config,constants
from localstack.utils.common import(chmod_r,clone,cp_r,disk_usage,download,format_number,is_command_available,load_file,mkdir,new_tmp_dir,new_tmp_file,retry,rm_rf,run,safe_requests,save_file,to_bytes,to_str,unzip)
from localstack.utils.docker_utils import DOCKER_CLIENT
from localstack.utils.testutil import create_zip_file
from localstack_ext import config as ext_config
from localstack_ext.bootstrap.licensing import get_auth_headers
from localstack_ext.bootstrap.state_utils import API_STATES_DIR,api_states_traverse
from localstack_ext.constants import API_PATH_PODS
LOG=logging.getLogger(__name__)
PERSISTED_FOLDERS=["api_states","dynamodb","kinesis"]
MERGE_STRATEGY_TWO_WAY="two way"
MERGE_STRATEGY_THREE_WAY="three way"
MERGE_STRATEGY_DISABLED="disabled"
class PodInfo:
 def __init__(self,name=Lhtri,pod_size=0):
  self.name=name
  self.pod_size=pod_size
  self.pod_size_compressed=0
  self.persisted_resource_names=[]
class CloudPodManager(Lhtrp):
 BACKEND="_none_"
 def __init__(self,pod_name=Lhtri,config=Lhtri):
  self.pod_name=pod_name
  self._pod_config=config
 def init(self):
  raise LhtrK("Not implemented")
 def delete(self,remote:LhtrY)->LhtrY:
  raise LhtrK("Not implemented")
 def push(self,comment:Lhtrq=Lhtri,three_way:LhtrY=LhtrP)->PodInfo:
  raise LhtrK("Not implemented")
 def push_overwrite(self,version:Lhtrf,comment:Lhtrq=Lhtri):
  pass
 def pull(self,inject_version_state:LhtrY=LhtrP,reset_state_before:LhtrY=LhtrP,fetch_all:LhtrY=Lhtri):
  raise LhtrK("Not implemented")
 def commit(self,message:Lhtrq=Lhtri):
  raise LhtrK("Not implemented")
 def inject(self,version:Lhtrf,reset_state:LhtrY):
  raise LhtrK("Not implemented")
 def list_versions(self)->List[Lhtrq]:
  raise LhtrK("Not implemented")
 def version_info(self,version:Lhtrf):
  raise LhtrK("Not implemented")
 def version_metamodel(self,version:Lhtrf):
  raise LhtrK("Not implemented")
 def set_version(self,version:Lhtrf,inject_version_state:LhtrY,reset_state:LhtrY,commit_before:LhtrY):
  raise LhtrK("Not implemented")
 def list_version_commits(self,version:Lhtrf)->List[Lhtrq]:
  raise LhtrK("Not implemented")
 def get_commit_diff(self,version:Lhtrf,commit:Lhtrf)->Lhtrq:
  raise LhtrK("Not implemented")
 def register_remote(self,pod_name:Lhtrq,ci_pod:LhtrY)->LhtrY:
  raise LhtrK("Not implemented")
 def rename_pod(self,current_pod_name,new_pod_name)->LhtrY:
  raise LhtrK("Not implemented")
 def list_pods(self,fetch_remote:LhtrY)->Set[Lhtrq]:
  raise LhtrK("Not implemented")
 def restart_container(self):
  LOG.info("Restarting LocalStack instance with updated persistence state - this may take some time ...")
  data={"action":"restart"}
  url="%s/health"%config.get_edge_url()
  try:
   requests.post(url,data=json.dumps(data))
  except requests.exceptions.ConnectionError:
   pass
  def check_status():
   LOG.info("Waiting for LocalStack instance to be fully initialized ...")
   response=requests.get(url)
   content=json.loads(to_str(response.content))
   statuses=[v for k,v in content["services"].items()]
   assert LhtrA(statuses)=={"running"}
  retry(check_status,sleep=3,retries=10)
 @LhtrE
 def pod_config(self):
  return self._pod_config or PodConfigManager.pod_config(self.pod_name)
 @Lhtrd
 def get(cls,pod_name,pre_config=Lhtri):
  pod_config=pre_config if pre_config else PodConfigManager.pod_config(pod_name)
  backend=pod_config.get("backend")
  for clazz in cls.__subclasses__():
   if clazz.BACKEND==backend:
    return clazz(pod_name=pod_name,config=pod_config)
  raise LhtrK('Unable to find Cloud Pod manager implementation type "%s"'%backend)
 def deploy_pod_into_instance(self,pod_path):
  delete_pod_zip=LhtrP
  if os.path.isdir(pod_path):
   tmpdir=new_tmp_dir()
   for folder in PERSISTED_FOLDERS:
    src_folder=os.path.join(pod_path,folder)
    if not os.path.exists(src_folder):
     continue
    tgt_folder=os.path.join(tmpdir,folder)
    cp_r(src_folder,tgt_folder,rm_dest_on_conflict=LhtrF)
   pod_path=create_zip_file(tmpdir)
   rm_rf(tmpdir)
   delete_pod_zip=LhtrF
  zip_content=load_file(pod_path,mode="rb")
  url=get_pods_endpoint()
  result=requests.post(url,data=zip_content)
  if result.status_code>=400:
   raise LhtrK("Unable to restore pod state via local pods management API %s (code %s): %s"%(url,result.status_code,result.content))
  if delete_pod_zip:
   rm_rf(pod_path)
  else:
   return pod_path
 @LhtrM
 def get_state_zip_from_instance(get_content=LhtrP):
  url=f"{get_pods_endpoint()}/state"
  result=requests.get(url)
  if result.status_code>=400:
   raise LhtrK("Unable to get local pod state via management API %s (code %s): %s"%(url,result.status_code,result.content))
  if get_content:
   return result.content
  zip_file=f"{new_tmp_file()}.zip"
  save_file(zip_file,result.content)
  return zip_file
 def get_pod_info(self,pod_data_dir:Lhtrq=Lhtri):
  result=PodInfo(self.pod_name)
  if pod_data_dir:
   result.pod_size=disk_usage(pod_data_dir)
   result.persisted_resource_names=get_persisted_resource_names(pod_data_dir)
  return result
class CloudPodManagerCPVCS(CloudPodManager):
 BACKEND="cpvcs"
 @LhtrM
 def parse_pod_name_from_qualifying_name(qualifying_name:Lhtrq)->Lhtrq:
  return qualifying_name.split(PODS_NAMESPACE_DELIM,1)[1]
 @LhtrM
 def _prepare_archives_from_presigned_urls(content):
  zip_path_version_space=new_tmp_file()
  presigned_urls=content.get("presigned_urls")
  version_space_url=presigned_urls.get("presigned_version_space_url")
  download(url=version_space_url,path=zip_path_version_space)
  zip_paths_state_archives={}
  zip_paths_meta_archives={}
  meta_and_state_urls=presigned_urls.get("presigned_meta_state_urls")
  for version_no,meta_and_state_url in meta_and_state_urls.items():
   zip_path_meta_archive=new_tmp_file()
   zip_path_state_archive=new_tmp_file()
   meta_url=meta_and_state_url["meta"]
   state_url=meta_and_state_url["state"]
   download(meta_url,zip_path_meta_archive)
   download(state_url,zip_path_state_archive)
   zip_paths_meta_archives[version_no]=zip_path_meta_archive
   zip_paths_state_archives[version_no]=zip_path_state_archive
  return zip_path_version_space,zip_paths_meta_archives,zip_paths_state_archives
 @LhtrM
 def _get_max_version_for_pod_from_platform(pod_name:Lhtrq,auth_headers):
  url=CloudPodManagerCPVCS.create_platform_url(f"{pod_name}/info/max-version")
  response=safe_requests.get(url=url,headers=auth_headers)
  if response.status_code!=200:
   LOG.warning("Failed to get version information from platform... aborting")
   return
  content=json.loads(response.content)
  remote_max_ver=Lhtrf(content["max_ver"])
  return remote_max_ver
 @LhtrM
 def _add_state_files_func(**kwargs):
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  dir_name=kwargs.get("dir_name")
  file_name=kwargs.get("fname")
  region=kwargs.get("region")
  service_name=kwargs.get("service_name")
  cpvcs_api.create_state_file_from_fs(path=dir_name,file_name=file_name,service=service_name,region=region)
 def _upload_version_and_product_space(self,presigned_urls):
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  presigned_version_space_url=presigned_urls.get("presigned_version_space_url")
  version_space_archive=cpvcs_api.create_version_space_archive()
  with Lhtrj(version_space_archive,"rb")as version_space_content:
   self.upload_content(presigned_version_space_url,version_space_content.read())
  presigned_meta_state_urls=presigned_urls.get("presigned_meta_state_urls")
  rm_rf(version_space_archive)
  for version_no,urls in presigned_meta_state_urls.items():
   meta_presigned_url=urls["meta"]
   meta_archive=cpvcs_api.get_version_meta_archive(version_no)
   with Lhtrj(meta_archive,"rb")as meta_archive_content:
    self.upload_content(meta_presigned_url,meta_archive_content)
   state_presigned_url=urls["state"]
   state_archive=cpvcs_api.get_version_state_archive(version_no)
   with Lhtrj(state_archive,"rb")as state_archive_content:
    self.upload_content(state_presigned_url,state_archive_content)
 def _add_state_to_cpvs_store(self):
  from localstack_ext.bootstrap.cpvcs.utils.common import config_context
  if not config_context.is_initialized():
   LOG.debug("No CPVCS instance detected - Could not push")
   return
  zip_file=self.get_state_zip_from_instance()
  tmp_dir=new_tmp_dir()
  with ZipFile(zip_file,"r")as state_zip:
   state_zip.extractall(tmp_dir)
   api_states_path=os.path.join(tmp_dir,API_STATES_DIR)
   api_states_traverse(api_states_path=api_states_path,side_effect=CloudPodManagerCPVCS._add_state_files_func,mutables=Lhtri)
  rm_rf(zip_file)
  rm_rf(tmp_dir)
 def _pull_versions(self,auth_headers,required_versions:Lhtrq):
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  url=self.create_platform_url(f"{self.pod_name}?versions={required_versions}")
  response=safe_requests.get(url=url,headers=auth_headers)
  if response.status_code!=200:
   LOG.warning("Failed to pull requested versions from platform")
   return
  content=json.loads(response.content)
  archives=CloudPodManagerCPVCS._prepare_archives_from_presigned_urls(content)
  zip_path_version_space=archives[0]
  zip_paths_meta_archives=archives[1]
  zip_paths_state_archives=archives[2]
  cpvcs_api.merge_from_remote(version_space_archive=zip_path_version_space,meta_archives=zip_paths_meta_archives,state_archives=zip_paths_state_archives)
 def _clone_pod(self,auth_headers):
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  url=self.create_platform_url(f"{self.pod_name}/clone")
  response=safe_requests.get(url,headers=auth_headers)
  if response.status_code!=200:
   LOG.warning(f"Failed to clone requested pod {self.pod_name}: {response.content}")
   return
  content=json.loads(response.content)
  archives=CloudPodManagerCPVCS._prepare_archives_from_presigned_urls(content)
  zip_path_version_space=archives[0]
  zip_paths_meta_archives=archives[1]
  zip_paths_state_archives=archives[2]
  remote_info={"storage_uuid":content.get("storage_uuid"),"qualifying_name":content.get("pod_name")}
  pod_name=CloudPodManagerCPVCS.parse_pod_name_from_qualifying_name(remote_info["qualifying_name"])
  cpvcs_api.init_remote(pod_name=pod_name,version_space_archive=zip_path_version_space,meta_archives=zip_paths_meta_archives,state_archives=zip_paths_state_archives,remote_info=remote_info)
 def init(self):
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_api.init(pod_name=self.pod_name)
 def delete(self,remote:LhtrY)->LhtrY:
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_dir=cpvcs_api.config_context.cpvcs_root_dir
  pod_dir=os.path.join(cpvcs_dir,self.pod_name)
  if os.path.isdir(pod_dir):
   rm_rf(pod_dir)
   return LhtrF
  if remote:
   pass
  return LhtrP
 def _push_to_remote(self,url:Lhtrq):
  auth_headers=get_auth_headers()
  response=safe_requests.put(url=url,headers=auth_headers)
  if response.status_code!=200:
   LOG.warning("Failed to get presigned URLs to upload new version.. aborting")
   return
  content=json.loads(response.content)
  presigned_urls=content.get("presigned_urls")
  self._upload_version_and_product_space(presigned_urls)
 def push(self,comment:Lhtrq=Lhtri,three_way:LhtrY=LhtrP)->PodInfo:
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_api.set_pod_context(self.pod_name)
  self._add_state_to_cpvs_store()
  if cpvcs_api.is_remotely_managed():
   auth_headers=get_auth_headers()
   local_max_ver=cpvcs_api.get_max_version_no()
   remote_max_ver=CloudPodManagerCPVCS._get_max_version_for_pod_from_platform(pod_name=self.pod_name,auth_headers=auth_headers)
   if local_max_ver<remote_max_ver:
    self.pull()
   cpvcs_api.push(comment=comment)
   url=CloudPodManagerCPVCS.create_platform_url(f"push/{self.pod_name}?version={local_max_ver + 1}")
   self._push_to_remote(url=url)
  else:
   created_version=cpvcs_api.push(comment=comment)
   LOG.debug(f"Created new version: {created_version}")
  return PodInfo()
 def push_overwrite(self,version:Lhtrf,comment:Lhtrq=Lhtri):
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_api.set_pod_context(pod_name=self.pod_name)
  if version>cpvcs_api.get_max_version_no():
   LOG.warning(f"Version {version} does not exist")
   return LhtrP
  self._add_state_to_cpvs_store()
  cpvcs_api.push_overwrite(version=version,comment=comment)
  if cpvcs_api.is_remotely_managed():
   url=CloudPodManagerCPVCS.create_platform_url(f"push-overwrite/{self.pod_name}?version={version}")
   self._push_to_remote(url=url)
  return LhtrF
 def pull(self,inject_version_state:LhtrY=LhtrP,reset_state_before:LhtrY=LhtrP,fetch_all:LhtrY=LhtrF):
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  auth_headers=get_auth_headers()
  if self.pod_name in cpvcs_api.list_locally_available_pods(show_remote_or_local=LhtrP):
   cpvcs_api.set_pod_context(self.pod_name)
   remote_max_ver=CloudPodManagerCPVCS._get_max_version_for_pod_from_platform(self.pod_name,auth_headers)
   if not remote_max_ver:
    return
   current_max_ver=cpvcs_api.get_max_version_no()
   if remote_max_ver==current_max_ver:
    LOG.info("No new version available remotely. Nothing to pull")
    return
   if fetch_all:
    required_versions=",".join(LhtrC(lambda ver:Lhtrq(ver),LhtrV(current_max_ver+1,remote_max_ver+1)))
   else:
    required_versions=current_max_ver
   self._pull_versions(auth_headers=auth_headers,required_versions=required_versions)
  else:
   self._clone_pod(auth_headers=auth_headers)
 def commit(self,message:Lhtrq=Lhtri):
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_api.set_pod_context(self.pod_name)
  self._add_state_to_cpvs_store()
  completed_revision=cpvcs_api.commit(message=message)
  LOG.debug(f"Completed revision: {completed_revision}")
 def inject(self,version:Lhtrf,reset_state:LhtrY):
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_api.set_pod_context(self.pod_name)
  if version==-1:
   version=cpvcs_api.get_head().version_number
  tmp_pod_path=cpvcs_api.get_version_state_archive(version)
  if not tmp_pod_path:
   LOG.warning(f"Could not find state for pod with version {version}")
   return
  if reset_state:
   reset_local_state(reset_data_dir=LhtrF,exclude_from_reset=["dynamodb","kinesis","stepfunctions"])
  self.deploy_pod_into_instance(tmp_pod_path)
 def list_versions(self)->List[Lhtrq]:
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_api.set_pod_context(self.pod_name)
  version_list=cpvcs_api.list_versions()
  return version_list
 def version_info(self,version:Lhtrf):
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_api.set_pod_context(self.pod_name)
  if version==-1:
   version=cpvcs_api.get_max_version_no()
  version_info=cpvcs_api.get_version_info(version)
  return version_info
 def version_metamodel(self,version:Lhtrf):
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_api.set_pod_context(self.pod_name)
  if version==-1:
   version=cpvcs_api.get_max_version_no()
  return cpvcs_api.get_version_metamodel(version_no=version)
 def set_version(self,version:Lhtrf,inject_version_state:LhtrY,reset_state:LhtrY,commit_before:LhtrY):
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_api.set_pod_context(self.pod_name)
  version_exists=cpvcs_api.set_active_version(version_no=version,commit_before=commit_before)
  if not version_exists:
   LOG.warning(f"Could not find version {version}")
  if inject_version_state:
   self.inject(version=version,reset_state=reset_state)
 def list_version_commits(self,version:Lhtrf)->List[Lhtrq]:
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_api.set_pod_context(self.pod_name)
  commits=cpvcs_api.list_version_commits(version_no=version)
  return commits
 def get_commit_diff(self,version:Lhtrf,commit:Lhtrf)->Lhtrq:
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_api.set_pod_context(self.pod_name)
  commit_diff=cpvcs_api.get_commit_diff(version_no=version,commit_no=commit)
  return commit_diff
 def register_remote(self,pod_name:Lhtrq,ci_pod:LhtrY=LhtrP)->LhtrY:
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  from localstack_ext.bootstrap.cpvcs.utils.remote_utils import register_remote
  cpvcs_api.set_pod_context(self.pod_name)
  max_ver=cpvcs_api.get_max_version_no()
  if max_ver==0:
   cpvcs_api.push("Init Version")
   max_ver=1
  auth_headers=get_auth_headers()
  url=self.create_platform_url("register")
  data={"pod_name":self.pod_name,"max_ver":max_ver,"ci_pod":ci_pod}
  data=json.dumps(data)
  response=safe_requests.post(url,data,headers=auth_headers)
  content=json.loads(response.content)
  if response.status_code!=200:
   LOG.warning(f"Failed to register pod {self.pod_name}: {content}")
   return LhtrP
  remote_info={"storage_uuid":content.get("storage_uuid"),"qualifying_name":content.get("pod_name")}
  presigned_urls=content.get("presigned_urls")
  self._upload_version_and_product_space(presigned_urls)
  register_remote(remote_info=remote_info)
  return LhtrF
 def rename_pod(self,current_pod_name,new_pod_name)->LhtrY:
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  cpvcs_api.set_pod_context(current_pod_name)
  if new_pod_name in cpvcs_api.list_locally_available_pods():
   LOG.warning(f"{new_pod_name} already exists locally")
   return LhtrP
  if cpvcs_api.is_remotely_managed():
   auth_headers=get_auth_headers()
   url=self.create_platform_url(f"{current_pod_name}/rename")
   data={"new_pod_name":new_pod_name}
   data=json.dumps(data)
   response=safe_requests.put(url,data,headers=auth_headers)
   if response.status_code!=200:
    LOG.warning(f"Failed to rename {current_pod_name} to {new_pod_name}: {response.content}")
    return LhtrP
  cpvcs_api.rename_pod(new_pod_name)
  return LhtrF
 def list_pods(self,fetch_remote:LhtrY)->Set[Lhtrq]:
  from localstack_ext.bootstrap.cpvcs import cpvcs_api
  result=cpvcs_api.list_locally_available_pods()
  if fetch_remote:
   auth_headers=get_auth_headers()
   url=self.create_platform_url("pods")
   response=safe_requests.get(url,headers=auth_headers)
   content=json.loads(response.content)
   for remote_pod in content.get("registered_pods")or[]:
    result.add(f"remote/{remote_pod}")
  return result
 @Lhtrd
 def upload_content(cls,presigned_url:Lhtrq,zip_data_content):
  res=safe_requests.put(presigned_url,data=zip_data_content)
  if res.status_code>=400:
   raise LhtrK("Unable to upload pod state to S3 (code %s): %s"%(res.status_code,res.content))
  return res
 @LhtrM
 def create_platform_url(request:Lhtrq)->Lhtrq:
  base_url="%s/cpvcs"%constants.API_ENDPOINT
  return os.path.join(base_url,request)
class CloudPodManagerFilesystem(CloudPodManager):
 BACKEND="file"
 def push(self,comment:Lhtrq=Lhtri,three_way:LhtrY=LhtrP)->PodInfo:
  local_folder=self.target_folder()
  print('Pushing state of cloud pod "%s" to local folder: %s'%(self.pod_name,local_folder))
  mkdir(local_folder)
  zip_file=self.get_state_zip_from_instance()
  unzip(zip_file,local_folder)
  chmod_r(local_folder,0o777)
  result=self.get_pod_info(local_folder)
  print("Done.")
  return result
 def pull(self,inject_version_state:LhtrY=LhtrP,reset_state_before:LhtrY=LhtrP,fetch_all:LhtrY=Lhtri):
  local_folder=self.target_folder()
  if not os.path.exists(local_folder):
   print('WARN: Local path of cloud pod "%s" does not exist: %s'%(self.pod_name,local_folder))
   return
  print('Pulling state of cloud pod "%s" from local folder: %s'%(self.pod_name,local_folder))
  self.deploy_pod_into_instance(local_folder)
 def target_folder(self):
  local_folder=re.sub(r"^file://","",self.pod_config.get("url",""))
  return local_folder
class CloudPodManagerManaged(CloudPodManager):
 BACKEND="managed"
 def push(self,comment:Lhtrq=Lhtri,three_way:LhtrY=LhtrP)->PodInfo:
  zip_data_content=self.get_state_zip_from_instance(get_content=LhtrF)
  print('Pushing state of cloud pod "%s" to backend server (%s KB)'%(self.pod_name,format_number(Lhtrk(zip_data_content)/1000.0)))
  self.push_content(self.pod_name,zip_data_content)
  print("Done.")
  result=self.get_pod_info()
  result.pod_size_compressed=Lhtrk(zip_data_content)
  return result
 def pull(self,inject_version_state:LhtrY=LhtrP,reset_state_before:LhtrY=LhtrP,fetch_all:LhtrY=Lhtri):
  presigned_url=self.presigned_url(self.pod_name,"pull")
  print('Pulling state of cloud pod "%s" from managed storage'%self.pod_name)
  zip_path=new_tmp_file()
  download(presigned_url,zip_path)
  self.deploy_pod_into_instance(zip_path)
  rm_rf(zip_path)
 @LhtrM
 def presigned_url(pod_name:Lhtrq,mode:Lhtrq)->Lhtrq:
  data={"pod_name":pod_name,"mode":mode}
  data=json.dumps(data)
  auth_headers=get_auth_headers()
  url="%s/cloudpods/data"%constants.API_ENDPOINT
  if ext_config.SYNC_POD_VERSION:
   url=f"{url}?version={ext_config.SYNC_POD_VERSION}"
  response=safe_requests.post(url,data,headers=auth_headers)
  content=response.content
  if response.status_code>=400:
   raise LhtrK("Unable to get cloud pod presigned URL (code %s): %s"%(response.status_code,content))
  content=json.loads(to_str(content))
  return content["presignedURL"]
 @Lhtrd
 def push_content(cls,pod_name,zip_data_content):
  presigned_url=cls.presigned_url(pod_name,"push")
  res=safe_requests.put(presigned_url,data=zip_data_content)
  if res.status_code>=400:
   raise LhtrK("Unable to push pod state to API (code %s): %s"%(res.status_code,res.content))
  return res
class CloudPodManagerGit(CloudPodManager):
 BACKEND="git"
 def push(self,comment:Lhtrq=Lhtri,three_way:LhtrY=LhtrP):
  repo=self.local_repo()
  branch=to_bytes(self.pod_config.get("branch"))
  remote_location=self.pod_config.get("url")
  try:
   porcelain.pull(repo,remote_location,refspecs=branch)
  except LhtrK as e:
   if self.has_git_cli():
    run("cd %s; git checkout %s; git pull"%(to_str(branch),self.clone_dir))
   else:
    LOG.info("Unable to pull repo: %s %s",e,traceback.format_exc())
  zip_file=self.get_state_zip_from_instance()
  tmp_data_dir=new_tmp_dir()
  unzip(zip_file,tmp_data_dir)
  is_empty_repo=b"HEAD" not in repo or repo.refs.allkeys()=={b"HEAD"}
  if is_empty_repo:
   LOG.debug("Initializing empty repository %s"%self.clone_dir)
   init_file=os.path.join(self.clone_dir,".init")
   save_file(init_file,"")
   porcelain.add(repo,init_file)
   porcelain.commit(repo,message="Initial commit")
  if branch not in repo:
   porcelain.branch_create(repo,branch,force=LhtrF)
  self.switch_branch(branch)
  for folder in PERSISTED_FOLDERS:
   LOG.info("Copying persistence folder %s to local git repo %s"%(folder,self.clone_dir))
   src_folder=os.path.join(tmp_data_dir,folder)
   tgt_folder=os.path.join(self.clone_dir,folder)
   cp_r(src_folder,tgt_folder)
   files=tgt_folder
   if os.path.isdir(files):
    files=[os.path.join(root,f)for root,_,files in os.walk(tgt_folder)for f in files]
   if files:
    porcelain.add(repo,files)
  porcelain.commit(repo,message="Update cloud pod state")
  try:
   porcelain.push(repo,remote_location,branch)
  except LhtrK:
   if not self.has_git_cli():
    raise
   run("cd %s; git push origin %s"%(self.clone_dir,to_str(branch)))
  result=self.get_pod_info(tmp_data_dir)
  return result
 def pull(self,inject_version_state:LhtrY=LhtrP,reset_state_before:LhtrY=LhtrP,fetch_all:LhtrY=Lhtri):
  repo=self.local_repo()
  client,path=self.client()
  remote_refs=client.fetch(path,repo)
  branch=self.pod_config.get("branch")
  remote_ref=b"refs/heads/%s"%to_bytes(branch)
  if remote_ref not in remote_refs:
   raise LhtrK('Unable to find branch "%s" in remote git repo'%branch)
  remote_location=self.pod_config.get("url")
  self.switch_branch(branch)
  branch_ref=b"refs/heads/%s"%to_bytes(branch)
  from dulwich.errors import HangupException
  try:
   porcelain.pull(repo,remote_location,branch_ref)
  except HangupException:
   pass
  self.deploy_pod_into_instance(self.clone_dir)
 def client(self):
  client,path=get_transport_and_path_from_url(self.pod_config.get("url"))
  return client,path
 def local_repo(self):
  self.clone_dir=Lhtrx(self,"clone_dir",Lhtri)
  if not self.clone_dir:
   pod_dir_name=re.sub(r"(\s|/)+","",self.pod_name)
   self.clone_dir=os.path.join(config.dirs.tmp,"pods",pod_dir_name,"repo")
   mkdir(self.clone_dir)
   if not os.path.exists(os.path.join(self.clone_dir,".git")):
    porcelain.clone(self.pod_config.get("url"),self.clone_dir)
    self.switch_branch(self.pod_config.get("branch"))
  return Repo(self.clone_dir)
 def switch_branch(self,branch):
  repo=self.local_repo()
  if self.has_git_cli():
   return run("cd %s; git checkout %s"%(self.clone_dir,to_str(branch)))
  branch_ref=b"refs/heads/%s"%to_bytes(branch)
  if branch_ref not in repo.refs:
   branch_ref=b"refs/remotes/origin/%s"%to_bytes(branch)
  repo.reset_index(repo[branch_ref].tree)
  repo.refs.set_symbolic_ref(b"HEAD",branch_ref)
 def has_git_cli(self):
  return is_command_available("git")
class PodConfigManagerMeta(Lhtrc):
 def __getattr__(cls,attr):
  def _call(*args,**kwargs):
   result=Lhtri
   for manager in cls.CHAIN:
    try:
     tmp=Lhtrx(manager,attr)(*args,**kwargs)
     if tmp:
      if not result:
       result=tmp
      elif LhtrH(tmp,LhtrT)and LhtrH(result,LhtrT):
       result.extend(tmp)
    except LhtrK:
     if LOG.isEnabledFor(logging.DEBUG):
      LOG.exception("error during PodConfigManager call chain")
   if result is not Lhtri:
    return result
   raise LhtrK('Unable to run operation "%s" for local or remote configuration'%attr)
  return _call
class PodConfigManager(Lhtrp,metaclass=PodConfigManagerMeta):
 CHAIN=[]
 @Lhtrd
 def pod_config(cls,pod_name):
  pods=PodConfigManager.list_pods()
  pod_config=[pod for pod in pods if pod["pod_name"]==pod_name]
  if not pod_config:
   raise LhtrK('Unable to find config for pod named "%s"'%pod_name)
  return pod_config[0]
class PodConfigManagerLocal(Lhtrp):
 CONFIG_FILE=".localstack.yml"
 def list_pods(self):
  local_pods=self._load_config(safe=LhtrF).get("pods",{})
  local_pods=[{"pod_name":k,"state":"Local Only",**v}for k,v in local_pods.items()]
  existing_names=LhtrA([pod["pod_name"]for pod in local_pods])
  result=[pod for pod in local_pods if pod["pod_name"]not in existing_names]
  return result
 def store_pod_metadata(self,pod_name,metadata):
  pass
 def _load_config(self,safe=LhtrP):
  try:
   return yaml.safe_load(to_str(load_file(self.CONFIG_FILE)))
  except LhtrK:
   if safe:
    return{}
   raise LhtrK('Unable to find and parse config file "%s"'%self.CONFIG_FILE)
class PodConfigManagerRemote(Lhtrp):
 def list_pods(self):
  result=[]
  auth_headers=get_auth_headers()
  url="%s/cloudpods"%constants.API_ENDPOINT
  response=safe_requests.get(url,headers=auth_headers)
  content=response.content
  if response.status_code>=400:
   raise LhtrK("Unable to fetch list of pods from API (code %s): %s"%(response.status_code,content))
  remote_pods=json.loads(to_str(content)).get("cloudpods",[])
  remote_pods=[{"state":"Shared",**pod}for pod in remote_pods]
  result.extend(remote_pods)
  return result
 def store_pod_metadata(self,pod_name,metadata):
  auth_headers=get_auth_headers()
  metadata["pod_name"]=pod_name
  response=safe_requests.post("%s/cloudpods"%constants.API_ENDPOINT,json.dumps(metadata),headers=auth_headers)
  content=response.content
  if response.status_code>=400:
   raise LhtrK("Unable to store pod metadata in API (code %s): %s"%(response.status_code,content))
  return json.loads(to_str(content))
PodConfigManager.CHAIN.append(PodConfigManagerLocal())
PodConfigManager.CHAIN.append(PodConfigManagerRemote())
def init_cpvcs(pod_name:Lhtrq,pre_config:Dict[Lhtrq,Lhtrq],**kwargs):
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 backend.init()
def delete_pod(pod_name:Lhtrq,remote:LhtrY,pre_config:Dict[Lhtrq,Lhtrq])->LhtrY:
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 result=backend.delete(remote=remote)
 return result
def register_remote(pod_name:Lhtrq,pre_config:Dict[Lhtrq,Lhtrq],**kwargs)->LhtrY:
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 result=backend.register_remote(pod_name=pod_name,ci_pod=pre_config.get("ci_pod",LhtrP))
 return result
def rename_pod(current_pod_name:Lhtrq,new_pod_name:Lhtrq,pre_config:Dict[Lhtrq,Lhtrq],**kwargs):
 backend=CloudPodManager.get(pod_name=current_pod_name,pre_config=pre_config)
 result=backend.rename_pod(current_pod_name=current_pod_name,new_pod_name=new_pod_name)
 return result
def list_pods_cpvcs(remote:LhtrY,pre_config:Dict[Lhtrq,Lhtrq],**kwargs)->List[Lhtrq]:
 backend=CloudPodManager.get(pod_name="",pre_config=pre_config)
 result=backend.list_pods(fetch_remote=remote)
 return result
def commit_state(pod_name:Lhtrq,pre_config:Dict[Lhtrq,Lhtrq],message:Lhtrq=Lhtri,**kwargs):
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 backend.commit(message=message)
def inject_state(pod_name:Lhtrq,version:Lhtrf,reset_state:LhtrY,pre_config:Dict[Lhtrq,Lhtrq],**kwargs):
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 backend.inject(version=version,reset_state=reset_state)
def list_versions(pod_name:Lhtrq,pre_config:Dict[Lhtrq,Lhtrq],**kwargs)->List[Lhtrq]:
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 versions=backend.list_versions()
 return versions
def get_version_info(version:Lhtrf,pod_name:Lhtrq,pre_config:Dict[Lhtrq,Lhtrq],**kwargs):
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 info=backend.version_info(version=version)
 return info
def get_version_metamodel(version:Lhtrf,pod_name:Lhtrq,pre_config:Dict[Lhtrq,Lhtrq],**kwargs):
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 metamodel=backend.version_metamodel(version=version)
 return metamodel
def set_version(version:Lhtrf,inject_version_state:LhtrY,reset_state:LhtrY,commit_before:LhtrY,pod_name:Lhtrq,pre_config:Dict[Lhtrq,Lhtrq],**kwargs)->LhtrY:
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 success=backend.set_version(version=version,inject_version_state=inject_version_state,reset_state=reset_state,commit_before=commit_before)
 return success
def list_version_commits(version:Lhtrf,pod_name:Lhtrq,pre_config:Dict[Lhtrq,Lhtrq])->List[Lhtrq]:
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 commits=backend.list_version_commits(version=version)
 return commits
def get_commit_diff(version:Lhtrf,commit:Lhtrf,pod_name:Lhtrq,pre_config:Dict[Lhtrq,Lhtrq])->Lhtrq:
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 commit_diff=backend.get_commit_diff(version=version,commit=commit)
 return commit_diff
def push_overwrite(version:Lhtrf,pod_name:Lhtrq,comment:Lhtrq,pre_config:Dict[Lhtrq,Lhtrq]):
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 backend.push_overwrite(version=version,comment=comment)
def push_state(pod_name,pre_config=Lhtri,squash_commits=LhtrP,comment=Lhtri,three_way=LhtrP,**kwargs):
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 pod_config=clone(backend.pod_config)
 pod_info=backend.push(comment=comment,three_way=three_way)
 pod_config["size"]=pod_info.pod_size or pod_info.pod_size_compressed
 pod_config["available_resources"]=pod_info.persisted_resource_names
 return pod_config
def get_pods_endpoint():
 edge_url=config.get_edge_url()
 return f"{edge_url}{API_PATH_PODS}"
def pull_state(pod_name,inject_version_state=LhtrP,reset_state_before=LhtrP,**kwargs):
 pre_config=kwargs.get("pre_config",Lhtri)
 if not pod_name:
  raise LhtrK("Need to specify a pod name")
 backend=CloudPodManager.get(pod_name=pod_name,pre_config=pre_config)
 backend.pull(inject_version_state=inject_version_state,reset_state_before=reset_state_before)
 print("Done.")
def reset_local_state(reset_data_dir=LhtrP,exclude_from_reset:List[Lhtrq]=Lhtri):
 url=f"{get_pods_endpoint()}/state"
 if reset_data_dir:
  url+="/datadir"
 if exclude_from_reset:
  url+=f"?exclude={','.join(exclude_from_reset)}"
 print("Sending request to reset the service states in local instance ...")
 result=requests.delete(url)
 if result.status_code>=400:
  raise LhtrK("Unable to reset service state via local management API %s (code %s): %s"%(url,result.status_code,result.content))
 print("Done.")
def list_pods(args):
 return PodConfigManager.list_pods()
def get_data_dir_from_container()->Lhtrq:
 try:
  details=DOCKER_CLIENT.inspect_container(config.MAIN_CONTAINER_NAME)
  mounts=details.get("Mounts")
  env=details.get("Config",{}).get("Env",[])
  data_dir_env=[e for e in env if e.startswith("DATA_DIR=")][0].partition("=")[2]
  try:
   data_dir_host=[m for m in mounts if m["Destination"]==data_dir_env][0]["Source"]
   data_dir_host=re.sub(r"^(/host_mnt)?",r"",data_dir_host)
   data_dir_env=data_dir_host
  except LhtrK:
   LOG.debug(f"No docker volume for data dir '{data_dir_env}' detected")
  return data_dir_env
 except LhtrK:
  LOG.warning('''Unable to determine DATA_DIR from LocalStack Docker container - please make sure $MAIN_CONTAINER_NAME is configured properly''')
def get_persisted_resource_names(data_dir)->List[Lhtrq]:
 names=[]
 with os.scandir(data_dir)as entries:
  for entry in entries:
   if entry.is_dir()and entry.name!="api_states":
    names.append(entry.name)
 with os.scandir(os.path.join(data_dir,"api_states"))as entries:
  for entry in entries:
   if entry.is_dir()and Lhtrk(os.listdir(entry.path))>0:
    names.append(entry.name)
 LOG.debug(f"Detected state files for the following APIs: {names}")
 return names
PODS_NAMESPACE_DELIM="-"
# Created by pyminifier (https://github.com/liftoff/pyminifier)



import abc
mhotC=None
mhotM=str
mhotG=classmethod
mhotf=bool
mhotA=False
mhotY=True
mhotJ=isinstance
import logging
import os
from typing import Dict,List,Union
from localstack_ext.bootstrap.cpvcs.models import CPVCSObj,Revision,Version
from localstack_ext.bootstrap.cpvcs.utils.common import config_context
from localstack_ext.bootstrap.cpvcs.utils.serializers import CPVCSSerializer,txt_serializers
LOG=logging.getLogger(__name__)
class StateFileLocator(abc.ABC):
 location="_none_"
 active_instance=mhotC
 @abc.abstractmethod
 def get_state_file_location_by_key(self,key:mhotM)->mhotM:
  pass
 @mhotG
 def get(cls,requested_file_locator:mhotM):
  if not cls.active_instance or cls.active_instance.location!=requested_file_locator:
   for clazz in cls.__subclasses__():
    if clazz.location==requested_file_locator:
     cls.active_instance=clazz()
  return cls.active_instance
class StateFileLocatorLocal(StateFileLocator):
 location="local"
 def get_state_file_location_by_key(self,key:mhotM)->mhotM:
  return os.path.join(config_context.get_obj_store_path(),key)
class ObjectStorageProvider(abc.ABC):
 location="_none_"
 active_instance=mhotC
 @mhotG
 def get(cls,state_file_locator:mhotM,requested_storage:mhotM,serializers:Dict[mhotM,CPVCSSerializer]):
  if not cls.active_instance or cls.active_instance.location!=requested_storage:
   state_file_locator=StateFileLocator.get(requested_file_locator=state_file_locator)
   for clazz in cls.__subclasses__():
    if clazz.location==requested_storage:
     cls.active_instance=clazz(state_file_locator=state_file_locator,serializers=serializers)
  return cls.active_instance
 def __init__(self,state_file_locator:StateFileLocator,serializers:Dict[mhotM,CPVCSSerializer]):
  self.state_file_locator=state_file_locator
  self.serializers=serializers
 @abc.abstractmethod
 def get_terminal_revision(self,revision_path_root:Revision):
  pass
 @abc.abstractmethod
 def get_state_file_location_by_key(self,key:mhotM)->mhotM:
  pass
 @abc.abstractmethod
 def get_version_by_key(self,key:mhotM)->Version:
  pass
 @abc.abstractmethod
 def get_revision_by_key(self,key:mhotM)->Revision:
  pass
 @abc.abstractmethod
 def get_revision_or_version_by_key(self,key:mhotM)->Union[Revision,Version]:
  pass
 @abc.abstractmethod
 def get_delta_file_by_key(self,key:mhotM)->mhotM:
  pass
 @abc.abstractmethod
 def upsert_objects(self,*args:CPVCSObj)->List[mhotM]:
  pass
 @abc.abstractmethod
 def update_revision_key(self,old_key:mhotM,new_key:mhotM,referenced_by_version:mhotM=mhotC):
  pass
 @abc.abstractmethod
 def version_exists(self,key:mhotM)->mhotf:
  pass
 @abc.abstractmethod
 def merge_remote_into_local_version(self,remote_location:mhotM,key:mhotM):
  pass
 @abc.abstractmethod
 def _update_key(self,old_key:mhotM,new_key:mhotM,base_path:mhotM)->mhotf:
  pass
 @abc.abstractmethod
 def _serialize(self,*args:CPVCSObj)->List[mhotM]:
  pass
 @abc.abstractmethod
 def _deserialize(self,key_serializer:Dict[mhotM,mhotM],remote_location:mhotM=mhotC)->List[CPVCSObj]:
  pass
class ObjectStorageLocal(ObjectStorageProvider):
 location="local"
 def get_terminal_revision(self,revision_path_root:Revision)->Revision:
  result=revision_path_root
  while result.assoc_commit:
   result=self.get_revision_by_key(result.assoc_commit.head_ptr)
  return result
 def get_state_file_location_by_key(self,key:mhotM)->mhotM:
  return self.state_file_locator.get_state_file_location_by_key(key)
 def get_version_by_key(self,key:mhotM)->Version:
  return self._deserialize({key:"version"})[0]
 def get_revision_by_key(self,key:mhotM)->Revision:
  return self._deserialize({key:"revision"})[0]
 def get_revision_or_version_by_key(self,key:mhotM)->Union[Revision,Version]:
  revision=self._deserialize({key:"revision"})[0]
  if revision:
   return revision
  version=self._deserialize({key:"version"})[0]
  if version:
   return version
  LOG.warning(f"No revision or version found with key {key}")
 def get_delta_file_by_key(self,key:mhotM)->mhotM:
  path=os.path.join(config_context.get_delta_log_path(),key)
  if os.path.isfile(path):
   return path
  LOG.warning(f"No state file found for key: {key}")
 def upsert_objects(self,*args:CPVCSObj)->List[mhotM]:
  return self._serialize(*args)
 def update_revision_key(self,old_key:mhotM,new_key:mhotM,referenced_by_version:mhotM=mhotC):
  updated=self._update_key(old_key,new_key,config_context.get_rev_obj_store_path())
  if not updated:
   LOG.warning(f"No revision found with key {old_key} to update")
   return
  if referenced_by_version:
   version=self.get_version_by_key(referenced_by_version)
   version.active_revision_ptr=new_key
   version.outgoing_revision_ptrs.remove(old_key)
   version.outgoing_revision_ptrs.add(new_key)
   self.upsert_objects(version)
 def version_exists(self,key:mhotM)->mhotf:
  return os.path.isfile(os.path.join(config_context.get_ver_obj_store_path(),key))
 def merge_remote_into_local_version(self,remote_location:mhotM,key:mhotM):
  local_version=self.get_version_by_key(key)
  remote_version=self._deserialize({key:"version"},remote_location)[0]
  local_version.outgoing_revision_ptrs.update(remote_version.outgoing_revision_ptrs)
  self.upsert_objects(local_version)
 def _update_key(self,old_key:mhotM,new_key:mhotM,base_path:mhotM)->mhotf:
  path_old_key=os.path.join(base_path,old_key)
  if not os.path.isfile(path_old_key):
   return mhotA
  path_new_key=os.path.join(base_path,new_key)
  os.rename(path_old_key,path_new_key)
  return mhotY
 def _serialize(self,*args:CPVCSObj)->List[mhotM]:
  hash_refs=[]
  for cpvcs_obj in args:
   if mhotJ(cpvcs_obj,Version):
    hash_refs.append(self.serializers["version"].store_obj(cpvcs_obj))
   elif mhotJ(cpvcs_obj,Revision):
    hash_refs.append(self.serializers["revision"].store_obj(cpvcs_obj))
  return hash_refs
 def _deserialize(self,key_serializer:Dict[mhotM,mhotM],remote_location:mhotM=mhotC)->List[Union[Revision,Version]]:
  return[self.serializers[serializer].retrieve_obj(key,remote_location)for key,serializer in key_serializer.items()]
default_storage=ObjectStorageProvider.get(state_file_locator="local",requested_storage="local",serializers=txt_serializers)
# Created by pyminifier (https://github.com/liftoff/pyminifier)
